# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any

from qcodes.instrument import Instrument
from qcodes.metadatable import Metadatable
from qcodes.parameters import Parameter

logger = logging.getLogger(__name__)


def is_bufferable(object: Instrument | Parameter):
    """Checks if the instrument or parameter is bufferable using the qumada Buffer definition."""
    if isinstance(object, Parameter):
        object = object.root_instrument
    return hasattr(object, "_qumada_buffer")  # and isinstance(object._qumada_buffer, Buffer)
    # TODO: check, if parameter can really be buffered
    # TODO: For some reason checking the type doesn't work all the time. Check again later.


def is_triggerable(object: Instrument | Parameter):
    """
    Checks if the instrument or parameter can be triggered by checking the corresponding flag in
    the mapping
    """
    if isinstance(object, Parameter):
        object = object.root_instrument
    return object._is_triggerable


class BufferException(Exception):
    """General Buffer Exception"""


def map_buffers(
    components: Mapping[Any, Metadatable],
    skip_mapped=True,
    **kwargs,
) -> None:
    """
    Maps the bufferable instruments of gate parameters.

    Args:
        components (Mapping[Any, Metadatable]): Instruments/Components in QCoDeS
    """

    buffered_instruments = filter(is_bufferable, components.values())
    for instrument in buffered_instruments:
        buffer = instrument._qumada_buffer
        if skip_mapped:
            if buffer.trigger in buffer.AVAILABLE_TRIGGERS:
                return
        print("Available trigger inputs:")
        print("[0]: None")
        for idx, trigger in enumerate(buffer.AVAILABLE_TRIGGERS, 1):
            print(f"[{idx}]: {trigger}")
        chosen = int(input(f"Choose the trigger input for {instrument.name}: "))
        if chosen == 0:
            trigger = None
        else:
            trigger = buffer.AVAILABLE_TRIGGERS[chosen - 1]
        buffer.trigger = trigger
        print(f"{buffer.trigger=}")


def _map_triggers(
    components: Mapping[Any, Metadatable],
    skip_mapped=True,
    **kwargs,
) -> None:
    """
    Maps the bufferable instruments of gate parameters.

    Args:
        components (Mapping[Any, Metadatable]): Instruments/Components in QCoDeS
    """
    triggered_instruments = filter(is_triggerable, components.values())
    for instrument in triggered_instruments:
        if skip_mapped:
            if instrument._qumada_mapping.trigger_in in instrument._qumada_mapping.AVAILABLE_TRIGGERS:
                return
        print("Available trigger inputs:")
        print("[0]: None")
        for idx, trigger in enumerate(instrument._qumada_mapping.AVAILABLE_TRIGGERS, 1):
            print(f"[{idx}]: {trigger}")
        chosen = int(input(f"Choose the trigger input for {instrument.name}: "))
        if chosen == 0:
            trigger = None
        else:
            trigger = instrument._qumada_mapping.AVAILABLE_TRIGGERS[chosen - 1]
        instrument._qumada_mapping.trigger_in = trigger
        print(f"trigger input = {instrument._qumada_mapping.trigger_in}")


def map_triggers(
    components: Mapping[Any, Metadatable],
    skip_mapped=True,
    path: None | str = None,
    **kwargs,
) -> None:
    """
    Maps the triggers of triggerable or bufferable components.
    Ignores already mapped triggers by default.

    Parameters
    ----------
    components : Mapping[Any, Metadatable]
        Components of QCoDeS station (containing instruments to be mapped).
    properties : dict
        Properties of measurement script/device. Currently only required for
        buffered instruments.
        TODO: Remove!
    gate_parameters : Mapping[Any, Mapping[Any, Parameter] | Parameter]
        Parameters of measurement script/device. Currently only required for
        buffered instruments.
        TODO: Remove!
    skip_mapped : Bool, optional
        If true already mapped parameters are skipped
        Set to false if you want to remap something. The default is True.
    path : None|str, optional
        Provide path to a json file with trigger mapping. If not all instruments
        are covered in the file, you will be asked to map those. Works only if
        names in file match instrument.full_name of your current instruments.
        The default is None.
    """
    if path is not None:
        try:
            load_trigger_mapping(components, path)
        except Exception as e:
            logger.warning(f"Exception when loadig trigger mapping from file: {e}")
    map_buffers(
        components,
        skip_mapped,
        **kwargs,
    )
    _map_triggers(components, skip_mapped, **kwargs)


def save_trigger_mapping(components: Mapping[Any, Metadatable], path: str):
    """
    Saves mapped triggers from components to json file.
    Components should be station.components, path is the path to the file.
    """
    trigger_dict = {}
    triggered_instruments = filter(lambda x: any((is_triggerable(x), is_bufferable(x))), components.values())
    for instrument in triggered_instruments:
        try:
            trigger_dict[instrument.full_name] = instrument._qumada_mapping.trigger_in
        except AttributeError:
            trigger_dict[instrument.full_name] = instrument._qumada_buffer.trigger
    with open(path, mode="w") as file:
        json.dump(trigger_dict, file)


def load_trigger_mapping(components: Mapping[Any, Metadatable], path: str):
    """
    Loads json file with trigger mappings and tries to apply them to instruments
    in the components. Works only if the instruments have the same full_name
    as in the saved file!
    """
    with open(path) as file:
        trigger_dict: Mapping[str, Mapping[str, str] | str] = json.load(file)
    for instrument_name, trigger in trigger_dict.items():
        for instrument in components.values():
            if instrument.full_name == instrument_name:
                if is_bufferable(instrument) is True:
                    instrument._qumada_buffer.trigger = trigger
                elif is_triggerable(instrument) is True:
                    instrument._qumada_mapping.trigger_in = trigger


class Buffer(ABC):
    """Base class for a general buffer interface for an instrument."""

    SETTING_NAMES: set[str] = {
        "trigger_mode",
        "trigger_threshold",
        "delay",
        "num_points",
        "channel",  # TODO: Remove? Should be part of the mapping.
        "sampling_rate",
        "duration",
        "burst_duration",
        "grid_interpolation",
        "num_bursts",
    }

    TRIGGER_MODE_NAMES: list[str] = [
        "continuous",
        "edge",
        "tracking_edge",
        "pulse",
        "tracking_pulse",
        "digital",
    ]

    TRIGGER_MODE_POLARITY_NAMES: list[str] = [
        "positive",
        "negative",
        "both",
    ]

    GRID_INTERPOLATION_NAMES: list[str] = [
        "exact",
        "nearest",
        "linear",
    ]

    AVAILABLE_TRIGGERS: list[str] = []

    settings_schema = {
        "type": "object",
        "properties": {
            "trigger_mode": {"type": "string", "enum": TRIGGER_MODE_NAMES},
            "trigger_mode_polarity": {
                "type": "string",
                "enum": TRIGGER_MODE_POLARITY_NAMES,
            },
            "grid_interpolation": {"type": "string", "enum": GRID_INTERPOLATION_NAMES},
            "trigger_threshold": {"type": "number"},
            "delay": {"type": "number"},
            "num_points": {"type": "integer"},
            "channel": {"type": "integer"},
            "sampling_rate": {"type": "number"},
            "duration": {"type": "number"},
            "burst_duration": {"type": "number"},
            "num_bursts": {"type": "integer"},
        },
        "oneOf": [
            {
                "required": ["sampling_rate", "duration"],
                "not": {"required": ["num_points"]},
            },
            {
                "required": ["sampling_rate", "num_points"],
                "not": {"required": ["duration"]},
            },
            {
                "required": ["duration", "num_points"],
                "not": {"required": ["sampling_rate"]},
            },
        ],
        "additionalProperties": False,
    }

    @abstractmethod
    def setup_buffer(self, settings: dict) -> None:
        """Sets instrument related settings for the buffer."""

    @property  # type: ignore
    @abstractmethod
    def trigger(self) -> Parameter | None:
        """
        The parameter, that triggers the instruments buffer.
        Set the trigger parameter using a qcodes parameter.
        """

    @trigger.setter  # type: ignore
    @abstractmethod
    def trigger(self, parameter: Parameter | None) -> None: ...

    @property
    @abstractmethod
    def num_points(self) -> int | None:
        """
        Number of points to write into buffer for each burst.
        Required to setup qcodes datastructure and to compare with max. buffer length.
        """
        # TODO: Handle multiple bursts

    @num_points.setter
    @abstractmethod
    def num_points(self) -> None: ...

    @abstractmethod
    def force_trigger(self) -> None:
        """Triggers the trigger."""

    @abstractmethod
    def read(self) -> dict:
        """
        Read the buffer

        Output is a dict with the following structure:

        .. code-block:: python

            {
                "timestamps": list[float],
                "param1": list[float],
                "param2": list[float],
                ...
            }

        """

    @abstractmethod
    def read_raw(self) -> Any:
        """Read the buffer and return raw output."""

    @abstractmethod
    def subscribe(self, parameters: set | list[Parameter]) -> None:
        """Measure provided parameters with the buffer."""

    @abstractmethod
    def unsubscribe(self, parameters: set | list[Parameter]) -> None:
        """Unsubscribe provided parameters, if they were subscribed."""

    @abstractmethod
    def is_subscribed(self, parameter: Parameter) -> bool:
        """True, if the parameter is subscribed and saved in buffer."""

    @abstractmethod
    def start(self) -> None:
        """Start the buffer. This is not the trigger."""

    @abstractmethod
    def stop(self) -> None:
        """Stop the buffer."""

    @abstractmethod
    def is_ready(self) -> bool:
        """True, if buffer is correctly initialized and ready to measure."""

    @abstractmethod
    def is_finished(self) -> bool:
        """True, if measurement is done and data has finished reading from the buffer."""


# class SoftwareTrigger(Parameter):
#     def __init__(self, **kwargs):
#         self._triggers = []
#         super().__init__(**kwargs)

#     def add_trigger(self, callable: Callable):
#         self._triggers.append(callable)

# TODO: Put buffer classes in separate files? Might become a bit crowded here in the future...



from __future__ import annotations

import logging
from abc import ABC
from copy import deepcopy
from typing import Any

import numpy as np
from qcodes import Station
from qcodes.parameters import Parameter
from qcodes.validators.validators import Numbers

from qumada.instrument.buffers.buffer import map_triggers
from qumada.instrument.mapping import map_terminals_gui
from qumada.measurement.measurement import MeasurementScript, load_param_whitelist
from qumada.measurement.scripts import (
    Generic_1D_Hysteresis_buffered,
    Generic_1D_parallel_Sweep,
    Generic_1D_Sweep,
    Generic_1D_Sweep_buffered,
    Generic_2D_Sweep_buffered,
    Generic_nD_Sweep,
    Generic_Pulsed_Measurement,
    Generic_Pulsed_Repeated_Measurement,
    Timetrace,
    Timetrace_buffered,
)
from qumada.utils.ramp_parameter import ramp_or_set_parameter

logger = logging.getLogger(__name__)


class Terminal_Exists_Exception(Exception):
    pass


class Parameter_Exists_Exception(Exception):
    pass


class QumadaDevice:
    def __init__(
        self,
        namespace=None,
        station: Station | None = None,
    ):
        self.namespace = namespace
        self.terminals = {}
        self.instrument_parameters = {}
        self.station = station
        self.buffer_settings = {}
        self.buffer_script_setup = {}
        self.states = {}
        self.ramp: bool = True

    def add_terminal(self, terminal_name: str, type: str | None = None, terminal_data: dict | None = {}):
        if terminal_name not in self.terminals.keys():
            self.__dict__[terminal_name.replace(" ", "_")] = self.terminals[terminal_name] = Terminal(
                terminal_name, self, type
            )
        else:
            raise Terminal_Exists_Exception(f"Terminal {terminal_name} already exists. Please remove it first!")
        if self.namespace is not None:
            if terminal_name not in self.namespace.keys():
                # Adding to the global namespace
                self.namespace[terminal_name.replace(" ", "_")] = self.terminals[terminal_name]
                logger.warning(f"Added {terminal_name} to global namespace!")
            else:
                raise Terminal_Exists_Exception(
                    f"Terminal {terminal_name} already exists in global namespace. \
                        Please remove it first!"
                )

    def remove_terminal(self, terminal_name: str):
        if terminal_name in self.terminals.keys():
            del self.__dict__[terminal_name]
            del self.terminals[terminal_name]
            if terminal_name in self.namespace:
                del self.namespace[terminal_name]
        else:
            logger.warning(f"{terminal_name} does not exist and could not be deleted")

    def update_terminal_parameters(self):
        for terminal, mapping in self.instrument_parameters.items():
            for param in mapping.keys():
                self.terminals[terminal].update_terminal_parameter(param)

    def save_defaults(self, ramp=None, **kwargs):
        """
        Saves current values as default for all Terminals and their parameters
        """
        for terminal in self.terminals.values():
            for param in terminal.terminal_parameters.values():
                param.save_default()

    def save_state(self, name: str):
        """
        Saves current state (inclung types, limits etc) as entry in the tuning dict with name as key.
        """
        self.states[name] = self.save_to_dict(priorize_stored_value=False)

    def set_state(self, name: str, ramp=None, **kwargs):
        if ramp is None:
            ramp = self.ramp
        self.load_from_dict(self.states[name])
        self.set_stored_values(ramp=ramp, **kwargs)

    def set_stored_values(self, ramp=None, **kwargs):
        if ramp is None:
            ramp = self.ramp
        for terminal in self.terminals.values():
            for param in terminal.terminal_parameters.values():
                param.set_stored_value()

    def set_defaults(self, ramp=None, **kwargs):
        """
        Sets all Terminals and their parameters to their default values
        """
        if ramp is None:
            ramp = self.ramp
        for terminal in self.terminals.values():
            for param in terminal.terminal_parameters.values():
                param.set_default(ramp=ramp, **kwargs)

    def voltages(self):
        """
        Prints all paramters called voltage from all Terminals of the device.
        """

        for terminal in self.terminals.values():
            try:
                label = terminal.name
                voltage = terminal.voltage()
                print(f"{label} {voltage=}")
            except AttributeError:
                pass

    @staticmethod
    def create_from_dict(data: dict, station: Station | None = None, namespace=None):
        """
        Creates a QumadaDevice object from valid parameter dictionaries as used in Qumada measurement scripts.
        Be aware that the validity is not checked at the moment, so there might be unexpected exceptions!
        Parameter values are not set upon initialization for safety reason! They are stored in
        the _stored_values attribute.
        By default terminals are added to the namespace provided by the namespace argument.
        If you set namespace=globals() you can make the terminals available in global namespace.
        TODO: Remove make_terminals_global parameter and check if namespace is not None
        """
        device = QumadaDevice(station=station, namespace=namespace)
        for terminal_name, terminal_data in data.items():
            device.add_terminal(terminal_name, terminal_data=terminal_data)
            for parameter_name, properties in terminal_data.items():
                device.terminals[terminal_name].add_terminal_parameter(parameter_name, properties=properties)
        return device

    def load_from_dict(self, data: dict):
        """
        Adds terminals and corresponding parameters to an existing QumadaDevice.
        Values are not set automatically for safety reasons, they are stored in the _stored_value attribute.
        TODO: Check behaviour for existing terminals/parameters
        """
        device = self
        for terminal_name, terminal_data in data.items():
            try:
                device.add_terminal(terminal_name, terminal_data=terminal_data)
            except Terminal_Exists_Exception:
                pass
            for parameter_name, properties in terminal_data.items():
                try:
                    device.terminals[terminal_name].add_terminal_parameter(parameter_name, properties=properties)
                except Parameter_Exists_Exception:
                    device.terminals[terminal_name].terminal_parameters[parameter_name].properties = properties
                    device.terminals[terminal_name].terminal_parameters[parameter_name]._apply_properties()

        return device

    def save_to_dict(self, priorize_stored_value=False):
        """
        Returns a dict compatible with the qumada measurements scripts.
        Contains type, setpoints, delay, start, stop, num_points and value of the
        terminal parameters.
        For the value, by default the current value of the parameter is used (the parameter is called
        therefore). If the parameter is not callable (e.g. because no mapping was done so far), the
        _stored_value attribute is used.
        If priorize_stored_values is set to True, the _stored_value attribute will be used if available
        and the return value of the parameters callable only if _stored_value is not available (or None).
        None values will be always ignored, the value will not be set in this case.
        """
        return_dict = {}

        for terminal_name, terminal in self.terminals.items():
            return_dict[terminal_name] = {}
            for param_name, param in terminal.terminal_parameters.items():
                return_dict[terminal_name][param_name] = {}
                for attr_name in [
                    "type",
                    "setpoints",
                    "delay",
                    "start",
                    "stop",
                    "num_points",
                    "break_conditions",
                    "limits",
                    "group",
                    "leverarms",
                    "compensated_gates",
                ]:
                    if hasattr(param, attr_name):
                        return_dict[terminal.name][param.name][attr_name] = getattr(param, attr_name)
                if priorize_stored_value:
                    if hasattr(param, "_stored_value") and getattr(param, "_stored_value") is not None:
                        return_dict[terminal.name][param.name]["value"] = getattr(param, "_stored_value")
                    elif callable(param):
                        try:
                            if param() is not None:
                                return_dict[terminal.name][param.name]["value"] = param()
                        except Exception as e:
                            logger.exception(e)
                    else:
                        logger.warning(f"Couldn't find value for {terminal_name} {param_name}")
                else:
                    try:
                        if param() is not None:
                            return_dict[terminal.name][param.name]["value"] = param()
                        else:
                            raise Exception(f"Calling {param} return None. Trying to use stored value")
                    except Exception as e:
                        logger.exception(e)
                        if hasattr(param, "_stored_value") and getattr(param, "_stored_value") is not None:
                            return_dict[terminal.name][param.name]["value"] = getattr(param, "_stored_value")
                        else:
                            logger.warning(f"Couldn't find value for {terminal_name} {param_name}")
        return return_dict

    def mapping(self, instrument_parameters: None | dict = None):
        if instrument_parameters is None:
            instrument_parameters = self.instrument_parameters
        if not isinstance(self.station, Station):
            raise TypeError("No valid qcodes station found. Make sure you have set the station attribute correctly!")
        map_terminals_gui(self.station.components, self.instrument_parameters, instrument_parameters)
        self.update_terminal_parameters()

    def timetrace(
        self,
        duration: float,
        timestep: float = 1,
        name=None,
        metadata=None,
        station=None,
        buffered=False,
        buffer_settings: dict | None = None,
        priorize_stored_value=False,
    ):
        """
        Perform a time-trace measurement over a specified duration and timestep.
        Uses the current values of the parameters. Can be buffered.

        Parameters
        ----------
        duration : float
            Total duration of the time-trace measurement in seconds.
        timestep : float, optional
            Time interval between data points in seconds. Default is 1.
        name : str, optional
            Measurement name. Default is None.
        metadata : dict, optional
            Metadata for the measurement. Default is None.
        station : Station, optional
            Station object associated with the measurement. Default is the station of the instance.
        buffered : bool, optional
            If True, performs a buffered time-trace measurement. Default is False.
        buffer_settings : dict, optional
            Buffer settings for the measurement. Default is the instance's buffer settings.
        priorize_stored_value : bool, optional
            If True, prioritizes stored values in the setup. Default is False.

        Returns
        -------
        data : qcodes.dataset.data_set.DataSet
            The dataset containing the measurement results.

        Raises
        ------
        TypeError
            If the provided `station` is not of type `Station`.

        Notes
        -----
        - Adjusts buffer settings temporarily when `buffered` is True.
        - Uses `Timetrace_buffered` for buffered measurements and `Timetrace` for unbuffered measurements.
        """
        if station is None:
            station = self.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        if buffer_settings is None:
            buffer_settings = self.buffer_settings
        temp_buffer_settings = deepcopy(buffer_settings)
        if buffered is True:
            logger.warning("Temporarily modifying buffer settings to match function arguments.")
            temp_buffer_settings["sampling_rate"] = 1 / timestep
            temp_buffer_settings["duration"] = duration
            temp_buffer_settings["burst_duration"] = duration
            try:
                del temp_buffer_settings["num_points"]
                del temp_buffer_settings["num_bursts"]
            except KeyError as e:
                logger.warning(e)

            script = Timetrace_buffered()
        else:
            script = Timetrace()
        script.setup(
            self.save_to_dict(priorize_stored_value=priorize_stored_value),
            metadata=metadata,
            measurement_name=name,
            duration=duration,
            timestep=timestep,
            buffer_settings=temp_buffer_settings,
            **self.buffer_script_setup,
        )
        mapping = self.instrument_parameters
        map_terminals_gui(station.components, script.gate_parameters, mapping)
        if buffered is True:
            map_triggers(station.components)
        data = script.run()
        return data

    def sweep_2D(
        self,
        slow_param: Parameter,
        fast_param: Parameter,
        slow_param_range: float,
        fast_param_range: float,
        slow_num_points: int = 50,
        fast_num_points: int = 100,
        name=None,
        metadata=None,
        station=None,
        buffered=False,
        buffer_settings: dict | None = None,
        priorize_stored_value=False,
        restore_state=True,
    ):
        """
        Perform a 2D sweep over two parameters. The current values are in the
        center of the sweep (the sweep ranges from currentvalue - 0.5*range to
        current value + 0.5*range). Can be buffered.

        Parameters
        ----------
        slow_param : Parameter
            The slow parameter to be swept.
        fast_param : Parameter
            The fast parameter to be swept.
        slow_param_range : float
            Range for the slow parameter sweep.
        fast_param_range : float
            Range for the fast parameter sweep.
        slow_num_points : int, optional
            Number of points for the slow parameter sweep. Default is 50.
        fast_num_points : int, optional
            Number of points for the fast parameter sweep. Default is 100.
        name : str, optional
            Measurement name. Default is None.
        metadata : dict, optional
            Metadata for the measurement. Default is None.
        station : Station, optional
            Station object associated with the measurement. Default is the station of the instance.
        buffered : bool, optional
            If True, performs a buffered 2D sweep. Default is False.
        buffer_settings : dict, optional
            Buffer settings for the measurement. Default is the instance's buffer settings.
        priorize_stored_value : bool, optional
            If True, prioritizes stored values in the setup. Default is False.
        restore_state : bool, optional
            If True, restores the original state of the parameters after the measurement. Default is True.

        Returns
        -------
        data : qcodes.dataset.data_set.DataSet
            The dataset containing the measurement results.

        Raises
        ------
        TypeError
            If the provided `station` is not of type `Station`.
        Exception
            If buffer settings are invalid or a measurement error occurs.

        Notes
        -----
        - Uses `Generic_2D_Sweep_buffered` for buffered measurements and `Generic_nD_Sweep` for unbuffered measurements.
        - Temporarily modifies buffer settings if `buffered` is True.
        - Restores the parameter state upon completion or exception.
        """
        if station is None:
            station = self.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        self.save_state("_temp_2D")
        try:
            for terminal in self.terminals.values():
                for parameter in terminal.terminal_parameters.values():
                    if parameter.type == "dynamic":
                        parameter.type = "static gettable"
            slow_param.type = "dynamic"
            slow_param.setpoints = np.linspace(
                slow_param.value - slow_param_range / 2.0, slow_param.value + slow_param_range / 2.0, slow_num_points
            )
            slow_param.group = 1
            fast_param.type = "dynamic"
            fast_param.group = 2
            fast_param.setpoints = np.linspace(
                fast_param.value - fast_param_range / 2.0, fast_param.value + fast_param_range / 2.0, fast_num_points
            )
            if buffer_settings is None:
                buffer_settings = self.buffer_settings
            temp_buffer_settings = deepcopy(buffer_settings)
            if buffered is True:
                if "num_points" in temp_buffer_settings.keys():
                    temp_buffer_settings["num_points"] = fast_num_points
                    logger.warning(
                        f"Temporarily changed buffer settings to match the \
                        number of points specified {fast_num_points=}"
                    )
                else:
                    logger.warning(
                        "Num_points not specified in buffer settings! fast_num_points value is \
                        ignored and buffer settings are used to specify measurement!"
                    )

                script = Generic_2D_Sweep_buffered()
            else:
                script = Generic_nD_Sweep()
            script.setup(
                self.save_to_dict(priorize_stored_value=priorize_stored_value),
                metadata=metadata,
                measurement_name=name,
                buffer_settings=temp_buffer_settings,
                **self.buffer_script_setup,
            )
            mapping = self.instrument_parameters
            map_terminals_gui(station.components, script.gate_parameters, mapping)
            if buffered is True:
                map_triggers(station.components)
            data = script.run()
        except Exception as e:
            print(self.states["_temp_2D"])
            self.set_state("_temp_2D")
            raise e
        finally:
            print(self.states["_temp_2D"])
            self.set_state("_temp_2D")
            del self.states["_temp_2D"]
        return data

    def sweep_parallel(
        self,
        params: list[Parameter],
        setpoints: list[list[float]] | None = None,
        target_values: list[float] | None = None,
        num_points: int = 100,
        name=None,
        metadata=None,
        station=None,
        priorize_stored_value=False,
        **kwargs,
    ):
        """
        Sweep multiple parameters in parallel.
        Provide either setpoints or target_values. Setpoints have to have the same length for all parameters.
        If no setpoints are provided, the target_values will be used to create the setpoints. Ramps will start from
        the current value of the parameters then.
        Gettable parameters and break conditions will be set according to their state in the device object.
        You can pass backsweep_after_break as a kwarg. If set to True, the sweep will continue in the opposite
        direction after a break condition is reached.


        Parameters
        ----------
        params : list[Parameter]
            List of parameters to be swept.
        setpoints : list[list[float]], optional
            A list of setpoints for each parameter. Each sublist must have the same length.
        target_values : list[float], optional
            Target values for each parameter. Used to generate setpoints if `setpoints` is not provided.
        num_points : int, optional
            Number of points for the generated setpoints. Default is 100.
        name : str, optional
            Measurement name. Default is None.
        metadata : dict, optional
            Metadata for the measurement. Default is None.
        station : Station, optional
            Station object associated with the measurement. Default is the station of the instance.
        priorize_stored_value : bool, optional
            If True, prioritizes stored values in the setup. Default is False.
        **kwargs
            Additional keyword arguments passed to the measurement script.

        Returns
        -------
        data : qcodes.dataset.data_set.DataSet
            The dataset containing the measurement results.

        Raises
        ------
        TypeError
            If the provided `station` is not of type `Station`.
        Exception
            If neither `setpoints` nor `target_values` are provided or both are provided.
        AssertionError
            If parameter or setpoint mismatches occur.

        Notes
        -----
        - Dynamic and static parameters are automatically configured during the measurement.
        - The script used for the measurement is `Generic_1D_parallel_Sweep`.

        """
        if station is None:
            station = self.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        if setpoints is None and target_values is None:
            raise (Exception("Either setpoints or target_values have to be provided!"))
        if target_values is not None and setpoints is not None:
            raise (Exception("Either setpoints or target_values have to be provided, not both!"))
        if setpoints is None:
            assert len(params) == len(target_values)
            setpoints = [np.linspace(param(), target, num_points) for param, target in zip(params, target_values)]
        assert len(params) == len(setpoints)
        assert all([len(setpoint) == len(setpoints[0]) for setpoint in setpoints])

        for terminal in self.terminals.values():
            for parameter in terminal.terminal_parameters.values():
                if parameter not in params and parameter.type == "dynamic":
                    parameter.type = "static gettable"
                if parameter in params:
                    parameter.type = "dynamic"
                    parameter.setpoints = setpoints[params.index(parameter)]
        script = Generic_1D_parallel_Sweep()
        script.setup(
            self.save_to_dict(priorize_stored_value=priorize_stored_value),
            metadata=metadata,
            measurement_name=name,
            **kwargs,
        )
        mapping = self.instrument_parameters
        map_terminals_gui(station.components, script.gate_parameters, mapping)
        data = script.run()
        return data

    def pulsed_measurement(
        self,
        params: list[Parameter],
        setpoints: list[list[float]],
        repetitions: int = 1,
        name=None,
        metadata=None,
        station=None,
        buffer_settings: dict | None = None,
        priorize_stored_value=False,
        **kwargs,
    ):
        """
        Perform a buffered pulsed measurement with optional repetitions.
        Results from repetitions are averaged.

        Parameters
        ----------
        params : list[Parameter]
            List of parameters to be pulsed.
        setpoints : list[list[float]]
            A list of setpoints for each parameter. Each sublist must have the same length.
        repetitions : int, optional
            Number of repetitions for the measurement. Default is 1.
        name : str, optional
            Measurement name. Default is None.
        metadata : dict, optional
            Metadata for the measurement. Default is None.
        station : Station, optional
            Station object associated with the measurement. Default is the station of the instance.
        buffer_settings : dict, optional
            Buffer settings for the measurement. Must include "num_points". Default is the instance's buffer settings.
        priorize_stored_value : bool, optional
            If True, prioritizes stored values in the setup. Default is False.
        **kwargs
            Additional keyword arguments passed to the measurement script.

        Returns
        -------
        data : qcodes.dataset.data_set.DataSet
            The dataset containing the measurement results.

        Raises
        ------
        TypeError
            If the provided `station` is not of type `Station`.
        AssertionError
            If parameter or setpoint mismatches occur.
        Exception
            If buffer settings are invalid.

        Notes
        -----
        - Configures dynamic and static parameters based on their usage in the measurement.
        - Uses `Generic_Pulsed_Measurement` for single repetition or
          `Generic_Pulsed_Repeated_Measurement` for multiple repetitions.
        - Buffer settings are adjusted to match the length of the setpoints.
        - Is always buffered (no need for buffered = True here)
        """
        if station is None:
            station = self.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        assert len(params) == len(setpoints)
        assert all([len(setpoint) == len(setpoints[0]) for setpoint in setpoints])
        assert repetitions >= 1
        if buffer_settings is None:
            buffer_settings = self.buffer_settings
        temp_buffer_settings = deepcopy(buffer_settings)

        if "num_points" in temp_buffer_settings.keys():
            temp_buffer_settings["num_points"] = len(setpoints[0])
            logger.warning(
                "Temporarily changed buffer settings to match the \
                number of points specified in the setpoints"
            )
        else:
            raise Exception(
                "For this kind of measurement, you have to specify the number of points in the buffer settings!"
            )
        for terminal in self.terminals.values():
            for parameter in terminal.terminal_parameters.values():
                if parameter not in params and parameter.type == "dynamic":
                    parameter.type = "static gettable"
                if parameter in params:
                    parameter.type = "dynamic"
                    parameter.setpoints = setpoints[params.index(parameter)]
        if repetitions == 1:
            script = Generic_Pulsed_Measurement()
        elif repetitions > 1:
            script = Generic_Pulsed_Repeated_Measurement()
        script.setup(
            self.save_to_dict(priorize_stored_value=priorize_stored_value),
            metadata=metadata,
            measurement_name=name,  # achtung geändert!
            repetitions=repetitions,
            buffer_settings=temp_buffer_settings,
            **self.buffer_script_setup,
            **kwargs,
        )
        mapping = self.instrument_parameters
        map_terminals_gui(station.components, script.gate_parameters, mapping)
        map_triggers(station.components, script.properties, script.gate_parameters)
        data = script.run()
        return data

    def run_measurement(
        self,
        script: MeasurementScript,
        dynamic_params: list,
        setpoints: list,
        dynamic_values: list | None = None,
        static_params: list | None = None,
        static_values: list | None = None,
        gettable_params: list | None = None,
        break_conditions: list | None = None,
        name=None,
        metadata=None,
        station=None,
        buffered=False,
        buffer_settings: dict | None = None,
        priorize_stored_value=False,
        **kwargs,
    ):
        """
        Runs any Qumada Measurement Script. Alters parameter attributes of device
        according to their type.


        Parameters
        ----------
        script : MeasurementScript
            The script you want to use. Has to be of type MeasurementScript.
        dynamic_params : list
            List of parameters that should be set to dynamic. Parameters that
            already are of the type "dynamic", but not listed here, will be set
            to "static gettable" to avoid user errors.
        setpoints : list
            Setpoints for the parameters listed before.
        dynamic_values : list|None, optional
            Values for the dynamic parameters (if required). Same length as dynamic
            parameters or None. The default is None.
        static_params : list|None, optional
            Parameters to set to static. Parameters that are already static will
            stay that way even if not listed here. For static gettable parameters
            add the parameter to the gettable list as well.The default is None.
        static_values : list|None, optional
            Values for the static params listed above explicitely. Same lenght as
            static params. The default is None.
        gettable_params : list|None, optional
            Parameters in this list are set to gettable. Parameters that are already
            gettable but not in this list, will stay gettable. The default is None.
        break_conditions : list|None, optional
            Break conditions for gettable parameters listed explicitely in the
            gettable_params above. Same lenght as gettable_params. The default is None.
        name : str|None, optional
            Custom name if required. If None name is generated by QuMada.
            The default is None.
        metadata : metadata object|None, optional
            Metadata Object to store. The default is None.
        station : QCoDeS Station, optional
            QCoDeS station. Overwrites device.station if not None. If None the
            device.station will be used. The default is None.
        buffered : Bool, optional
            Set to true for buffered measurements. Otherwise buffer settings etc
            are not passed on (measurements might still work though if everything
            was already defined in device. The default is False.
        buffer_settings : dict | None, optional
            Buffer settings. Overwrites settings stored in device if not None.
            The default is None.
        priorize_stored_value : Bool, optional
            Use values from dictionionary used to create device object instead
            of current values of the parameters if available. The default is False.

        **kwargs : dict|None
            Additional params, possibly depending on the measurement script.
            (e.g. duration for timetraces)


        Raises
        ------
        TypeError
            If no valid station available.
        Exception
            Read output, related to buffer settings.

        Returns
        -------
        data : list[Dataset]
            List of Datasets with measurement reuslts.

        """
        if station is None:
            station = self.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        assert len(dynamic_params) == len(setpoints)
        if dynamic_values is not None:
            assert len(dynamic_values) == len(dynamic_params)
        if static_values is not None:
            assert len(static_params) == len(static_values)
        if break_conditions is not None:
            assert len(gettable_params) == len(break_conditions)

        if buffer_settings is None:
            buffer_settings = self.buffer_settings
        temp_buffer_settings = deepcopy(buffer_settings)

        if "num_points" in temp_buffer_settings.keys():
            temp_buffer_settings["num_points"] = len(setpoints[0])
            logger.warning(
                "Temporarily changed buffer settings to match the \
                number of points specified in the setpoints"
            )
        else:
            raise Exception(
                "For this kind of measurement, you have to specify the number of points in the buffer settings!"
            )
        for terminal in self.terminals.values():
            for parameter in terminal.terminal_parameters.values():
                if dynamic_params is not None and parameter not in dynamic_params and parameter.type == "dynamic":
                    parameter.type = "static gettable"
                if dynamic_params is not None and parameter in dynamic_params:
                    parameter.type = "dynamic"
                    parameter.setpoints = setpoints[dynamic_params.index(parameter)]
                    if dynamic_values is not None:
                        parameter.value = dynamic_values[dynamic_params.index(parameter)]
                if static_params is not None and parameter in static_params:
                    parameter.type = "static"
                    if gettable_params is not None and parameter in gettable_params:
                        parameter.type = "static gettable"
                    if static_values is not None:
                        parameter.value = static_values[static_params.index(parameter)]
                elif gettable_params is not None and parameter in gettable_params:
                    parameter.type = "gettable"
                    if break_conditions is not None:
                        parameter.break_conditions = break_conditions[gettable_params.index(parameter)]

        script = script()
        script.setup(
            self.save_to_dict(priorize_stored_value=priorize_stored_value),
            metadata=metadata,
            measurement_name=name,
            buffer_settings=temp_buffer_settings,
            **self.buffer_script_setup,
            **kwargs,
        )
        mapping = self.instrument_parameters
        map_terminals_gui(station.components, script.gate_parameters, mapping)
        if buffered is True:
            map_triggers(station.components)
        data = script.run()
        return data


class Terminal(ABC):
    """
    Base class for Terminals scripts.

    The abstract functions "reset" has to be implemented.
    """

    PARAMETER_NAMES: set[str] = load_param_whitelist()

    def __init__(self, name, parent: QumadaDevice | None = None, type: str | None = None):

        self.properties: dict[Any, Any] = {}
        self.name = name
        self._parent = parent
        self.type = type
        self.terminal_parameters: dict[Any, dict[Any, Parameter | None] | Parameter | None] = {}

    def add_terminal_parameter(
        self, parameter_name: str, parameter: Parameter = None, properties: dict | None = None
    ) -> None:
        """
        Adds a gate parameter to self.terminal_parameters.

        Args:
            parameter_name (str): Name of the parameter. Has to be in MeasurementScript.PARAMETER_NAMES.
            terminal_name (str): Name of the parameter's gate. Set this, if you want to define the parameter
                             under a specific gate. Defaults to None.
            parameter (Parameter): Custom parameter. Set this, if you want to set a custom parameter. Defaults to None.
        """
        if parameter_name not in Terminal.PARAMETER_NAMES:
            raise NameError(f'parameter_name "{parameter_name}" not in MeasurementScript.PARAMETER_NAMES.')
        if parameter_name not in self.terminal_parameters.keys():
            self.__dict__[parameter_name] = self.terminal_parameters[parameter_name] = Terminal_Parameter(
                parameter_name, self, properties=properties
            )
            if self.name not in self._parent.instrument_parameters.keys():
                self._parent.instrument_parameters[self.name] = {}
            self._parent.instrument_parameters[self.name][parameter_name] = parameter
        else:
            raise Parameter_Exists_Exception(f"Parameter{parameter_name} already exists")

    def remove_terminal_parameter(self, parameter_name: str) -> None:
        """
        Adds a gate parameter to self.terminal_parameters.

        Args:
            parameter_name (str): Name of the parameter. Has to be in MeasurementScript.PARAMETER_NAMES.
            terminal_name (str): Name of the parameter's gate. Set this, if you want to define the parameter
                             under a specific gate. Defaults to None.
            parameter (Parameter): Custom parameter. Set this, if you want to set a custom parameter. Defaults to None.
        """
        if parameter_name in self.terminal_parameters.keys():
            del self.__dict__[parameter_name]
            del self.terminal_parameters[parameter_name]
        else:
            raise Exception(f"Parameter{parameter_name} does not exist!")

    def update_terminal_parameter(self, parameter_name: str, parameter: Parameter | None = None) -> None:
        self.terminal_parameters[parameter_name].instrument_parameter = self._parent.instrument_parameters[self.name][
            parameter_name
        ]

    def __call__(self, value=None):
        if "voltage" in self.terminal_parameters.keys():
            return self.voltage(value)
        else:
            raise TypeError


class Terminal_Parameter(ABC):
    def __init__(self, name: str, Terminal: Terminal, properties: dict = {}) -> None:
        self._parent = Terminal
        self._parent_device = Terminal._parent
        if properties is None:
            properties = {}
        self.properties: dict[Any, Any] = properties
        self.type = self.properties.get("type", None)
        self._stored_value = self.properties.get("value", None)  # For storing values for measurements
        self.setpoints = self.properties.get("setpoints", None)
        self.delay = self.properties.get("delay", 0)
        self.break_conditions = self.properties.get("break_conditions", [])
        self._value = None
        self.name = name
        self._limits = self.properties.get("limits", None)
        self.leverarms = self.properties.get("leverarms", None)
        self.compensated_gates = self.properties.get("compensated_gates")
        self.rampable = False
        self.ramp_rate = self.properties.get("ramp_rate", 0.1)
        self.group = self.properties.get("group", None)
        self.default_value = None
        self.scaling = 1  # Only relevant for setting values. Not taken into account for measurements!
        self._instrument_parameter = None
        self.locked = False
        self._limit_validator = None

    def reset(self):
        pass

    def _apply_properties(self):
        """
        Make sure changes to the properties are passed on to the object attributes
        """
        self.type = self.properties.get("type", self.type)
        self._stored_value = self.properties.get("value", self._stored_value)  # For storing values for measurements
        self.setpoints = self.properties.get("setpoints", self.setpoints)
        self.delay = self.properties.get("delay", self.delay)
        self.ramp_rate = self.properties.get("ramp_rate", self.ramp_rate)
        self.group = self.properties.get("group", self.group)
        self.leverarms = self.properties.get("leverarms", self.leverarms)
        self.compensated_gates = self.properties.get("compensated_gates")

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.locked is True:
            raise Exception(f"Parameter {self.name} of Terminal {self._parent.name} is locked and cannot be set!")
            return

        if isinstance(value, float):
            self._value = self.scaling * value
            try:
                self.instrument_parameter(self.scaling * value)
            except TypeError:
                self._parent_device.update_terminal_parameters()
                self.instrument_parameter(self.scaling * value)
        else:
            self._value = value
            # TODO: Replace Try/Except block, update_terminal_parameters() should be called by mapping function
            try:
                self.instrument_parameter(value)
            except TypeError:
                self._parent_device.update_terminal_parameters()
                self.instrument_parameter(value)

    @value.getter
    def value(self):
        # TODO: Replace Try/Except block, update_terminal_parameters() should be called by mapping function
        try:
            return self.instrument_parameter()
        except TypeError:
            self._parent_device.update_terminal_parameters()
            return self.instrument_parameter()

    @property
    def instrument_parameter(self):
        return self._instrument_parameter

    @instrument_parameter.setter
    def instrument_parameter(self, param: Parameter):
        if isinstance(param, Parameter) or param is None:
            self._instrument_parameter = param
            self._set_limits()
        else:
            raise TypeError(f"{param} is not a QCoDeS parameter!")

    @property
    def limits(self):
        return self._limits

    @limits.setter
    def limits(self, limits):
        if type(limits) in (list, tuple) and len(limits) == 2:
            self._limits = limits
            self._set_limits()
        else:
            raise ValueError("Limits has to be a list|tuple with two entries")

    def _set_limits(self):
        """
        Uses QCoDeS parameter's validators to limit values of parameters with
        number value to the values set in the limits attribute.
        Will replace last validator of corresponding parameter, if it was set
        by this method before! Won't remove validators that existed before
        initialization of the parameter. Make sure not to add validators
        manually to avoid problem (QCoDeS can only remove the last added
        validator, so it's not possible to just remove the correct one.')
        """
        if self.limits is None:
            return
        if len(self.limits) != 2:
            raise ValueError(f"Invalid limits provided for {self._parent.name} {self.name}")
        param = self.instrument_parameter
        if not isinstance(param, Parameter):
            logger.exception(
                f"Cannot set limits to {self._parent.name} {self.name} \
                             as no valid instrument parameter was assigned to it!"
            )
        else:
            try:
                if self._limit_validator in param.validators:
                    param.remove_validator()
            except AttributeError as e:
                logger.warning(e)
                pass
            self._limit_validator = Numbers(min_value=min(self.limits), max_value=max(self.limits))
            param.add_validator(self._limit_validator)

    def ramp(self, value, ramp_rate: float | None = None, ramp_time: float = 5, setpoint_intervall: float = 0.01):
        if ramp_rate is None:
            ramp_rate = self.ramp_rate
        ramp_or_set_parameter(
            self.instrument_parameter,
            value,
            ramp_rate=ramp_rate,
            ramp_time=ramp_time,
            setpoint_intervall=setpoint_intervall,
        )

    def measured_ramp(
        self,
        value,
        num_points=100,
        start=None,
        station=None,
        name=None,
        metadata=None,
        backsweep=False,
        buffered=False,
        buffer_settings: dict | None = None,
        priorize_stored_value=False,
    ):
        """
        Perform a ramp of the parameter value and measure all gettable parameters.
        Can be buffered.

        Parameters
        ----------
        value : float
            Target value for the ramp.
        num_points : int, optional
            Number of points for the ramp. Default is 100.
        start : float, optional
            Starting value for the ramp. If None, the current parameter value is used. Default is None.
        station : Station, optional
            Station object for the measurement. Default is the station of the parent device.
        name : str, optional
            Measurement name. Default is None.
        metadata : dict, optional
            Metadata for the measurement. Default is None.
        backsweep : bool, optional
            If True, includes a backsweep to return to the starting value after
            reaching the target value. Default is False.
        buffered : bool, optional
            If True, performs a buffered ramp measurement. Default is False.
        buffer_settings : dict, optional
            Additional buffer settings for the measurement. Default is None.
        priorize_stored_value : bool, optional
            If True, prioritizes stored values in the setup. Default is False.

        Returns
        -------
        data : qcodes.dataset.data_set.DataSet
            The dataset containing the measurement results.

        Raises
        ------
        TypeError
            If the provided `station` is not of type `Station`.
        Exception
            If the parameter is locked or invalid buffer settings are provided.

        Notes
        -----
        - Uses `Generic_1D_Hysteresis_buffered` for buffered ramps with backsweep.
        - Uses `Generic_1D_Sweep_buffered` for buffered ramps without backsweep.
        - Uses `Generic_1D_Sweep` for unbuffered ramps.
        - Temporarily modifies buffer settings to match the number of points if buffered.
        - Ensures all other dynamic parameters are set to "static gettable" before the ramp.
        """
        if station is None:
            station = self._parent_device.station
        if not isinstance(station, Station):
            raise TypeError("No valid station assigned!")
        if self.locked:
            raise Exception(f"{self.name} is locked!")

        for terminal_name, terminal in self._parent_device.terminals.items():
            for param_name, param in terminal.terminal_parameters.items():
                if param.type == "dynamic":
                    param.type = "static gettable"
        self.type = "dynamic"
        if start is None:
            start = self()
        if backsweep is True:
            if buffered is False:
                self.setpoints = [*np.linspace(start, value, num_points), *np.linspace(value, start, num_points)]
            else:
                self.setpoints = np.linspace(start, value, num_points)
        else:
            self.setpoints = np.linspace(start, value, num_points)
        temp_buffer_settings = deepcopy(self._parent_device.buffer_settings)
        temp_buffer_settings.update(buffer_settings or {})
        if buffered:
            if "num_points" in temp_buffer_settings.keys():
                temp_buffer_settings["num_points"] = num_points
                logger.warning(
                    f"Temporarily changed buffer settings to match the number of points specified {num_points=}"
                )
            else:
                logger.warning(
                    "Num_points not specified in buffer settings! fast_num_points value is \
                        ignored and buffer settings are used to specify measurement!"
                )
            if backsweep is True:
                script = Generic_1D_Hysteresis_buffered()
            else:
                script = Generic_1D_Sweep_buffered()
        else:
            script = Generic_1D_Sweep()
        script.setup(
            self._parent_device.save_to_dict(priorize_stored_value=priorize_stored_value),
            metadata=metadata,
            name=name,
            iterations=1,
            buffer_settings=temp_buffer_settings,
            **self._parent_device.buffer_script_setup,
        )
        mapping = self._parent_device.instrument_parameters
        map_terminals_gui(station.components, script.gate_parameters, mapping)
        if buffered is True:
            map_triggers(station.components, script.properties, script.gate_parameters)
        data = script.run()
        return data

    def save_default(self):
        """
        Saves current value as default value.
        """
        try:
            self.default_value = self.value
        except Exception as e:
            logger.warning(f"{e} was raised when trying to save default value of {self.name}")
            pass

    def set_default(self, ramp=True, **kwargs):
        """
        Sets value to default value
        """
        if self.default_value is not None:
            try:
                if ramp is True:
                    self.ramp(self.default_value, **kwargs)
                else:
                    self.value = self.default_value
            except NotImplementedError as e:
                logger.debug(f"{e} was raised and ignored")
        else:
            logger.warning(f"No default value set for parameter {self.name}")

    def set_stored_value(self, ramp=True, **kwargs):
        """
        Sets value to stored value from dict
        """
        if self._stored_value is not None:
            try:
                if ramp is True:
                    self.ramp(self._stored_value, **kwargs)
                else:
                    self.value = self._stored_value
            except NotImplementedError as e:
                logger.debug(f"{e} was raised and ignored")
        else:
            logger.warning(f"No stored value set for parameter {self.name}")

    def __call__(self, value=None, ramp=None):
        if value is None:
            return self.value
        else:
            if ramp is True:
                self.ramp(value)
            else:
                self.value = value


# class Virtual_Terminal_Parameter(Terminal_Parameter):



import abc
import itertools
import uuid
from collections.abc import Sequence
from time import sleep as sleep_s
from typing import NewType, Optional

import numpy as np
from packaging.version import parse
from pyvisa.errors import VisaIOError
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.instrument.visa import VisaInstrument
from qcodes.utils import validators

# Version 1.2.0
#
# Guiding principles for this driver for QDevil QDAC-II
# -----------------------------------------------------
#
# 1. Each command should be self-contained, so
#
#        qdac.ch02.dc_constant_V(0.1)
#
#    should make sure that channel 2 is in the right mode for outputting
#    a constant voltage.
#
# 2. Numeric values should be in ISO units and/or their unit should be an
#    explicitly part of the function name, like above.  If the numeric is
#    a unit-less number, then prefixed by n_ like
#
#        qdac.n_channels()
#
# 3. Allocation of resources should be automated as much as possible, preferably
#    by python context managers that automatically clean up on exit.  Such
#    context managers have a name with a '_Context' suffix.
#
# 4. Any generator should by default be set to start on the BUS trigger
#    (*TRG) so that it is possible to synchronise several generators without
#    further setup; which also eliminates the need for special cases for the
#    BUS trigger.


#
# Future improvements
# -------------------
#
# - Detect and handle mixing of internal and external triggers (_trigger).
#


error_ambiguous_wave = "Only one of frequency_Hz or period_s can be " "specified for a wave form"


def ints_to_comma_separated_list(array: Sequence[int]) -> str:
    return ",".join([str(x) for x in array])


def floats_to_comma_separated_list(array: Sequence[float]) -> str:
    rounded = [format(x, "g") for x in array]
    return ",".join(rounded)


def comma_sequence_to_list(sequence: str) -> Sequence[str]:
    if not sequence:
        return []
    return [x.strip() for x in sequence.split(",")]


def comma_sequence_to_list_of_floats(sequence: str) -> Sequence[float]:
    if not sequence:
        return []
    return [float(x.strip()) for x in sequence.split(",")]


def diff_matrix(initial: Sequence[float], measurements: Sequence[Sequence[float]]) -> np.ndarray:
    """Subtract an array of measurements by an initial measurement"""
    matrix = np.asarray(measurements)
    return matrix - np.asarray(list(itertools.repeat(initial, matrix.shape[1])))


def split_version_string_into_components(version: str) -> list[str]:
    return version.split("-")


# External input trigger

# There are four 3V3 non-isolated triggers on the back (1, 2, 3, 4).

ExternalInput = NewType("ExternalInput", int)


class QDac2Trigger_Context:
    """Internal Triggers with automatic deallocation

    This context manager wraps an already-allocated internal trigger number so
    that the trigger can be automatically reclaimed when the context exits.
    """

    def __init__(self, parent: "QDac2", value: int):
        self._parent = parent
        self._value = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._parent.free_trigger(self)
        # Propagate exceptions
        return False

    @property
    def value(self) -> int:
        """internal SCPI trigger number"""
        return self._value


def _trigger_context_to_value(trigger: QDac2Trigger_Context) -> int:
    return trigger.value


class QDac2ExternalTrigger(InstrumentChannel):
    """External output trigger

    There are three 5V isolated triggers on the front (1, 2, 3) and two
    non-isolated 3V3 on the back (4, 5).
    """

    def __init__(self, parent: "QDac2", name: str, external: int):
        super().__init__(parent, name)
        self.add_function(name="source_from_bus", call_cmd=f"outp:trig{external}:sour bus")
        self.add_parameter(
            name="source_from_input",
            # Route external input to external output
            set_cmd="outp:trig{}:sour ext{}".format(external, "{}"),
            get_parser=int,
        )
        self.add_parameter(
            name="source_from_trigger",
            # Route internal trigger to external output
            set_parser=_trigger_context_to_value,
            set_cmd="outp:trig{}:sour int{}".format(external, "{}"),
            get_parser=int,
        )
        self.add_parameter(
            name="width_s",
            label="width",
            unit="s",
            set_cmd="outp:trig{}:widt {}".format(external, "{}"),
            get_cmd=f"outp:trig{external}:widt?",
            get_parser=float,
        )
        self.add_parameter(
            name="polarity",
            label="polarity",
            set_cmd="outp:trig{}:pol {}".format(external, "{}"),
            get_cmd=f"outp:trig{external}:pol?",
            get_parser=str,
            vals=validators.Enum("inv", "norm"),
        )
        self.add_parameter(
            name="delay_s",
            label="delay",
            unit="s",
            set_cmd="outp:trig{}:del {}".format(external, "{}"),
            get_cmd=f"outp:trig{external}:del?",
            get_parser=float,
        )
        self.add_function(name="signal", call_cmd=f"outp:trig{external}:sign")
        self.add_parameter(
            name="set_trigger",
            # Route internal trigger to external output
            # set_parser=_trigger_context_to_value,
            set_cmd="outp:trig{}:sour {}".format(external, "{}"),
        )


class _Channel_Context(metaclass=abc.ABCMeta):

    def __init__(self, channel: "QDac2Channel"):
        self._channel = channel

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Propagate exceptions
        return False

    def allocate_trigger(self) -> QDac2Trigger_Context:
        """Allocate internal trigger

        Returns:
            QDac2Trigger_Context: Context that wraps the trigger
        """
        return self._channel._parent.allocate_trigger()

    @abc.abstractmethod
    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        pass

    @abc.abstractmethod
    def start_on_external(self, trigger: ExternalInput) -> None:
        pass

    @abc.abstractmethod
    def abort(self) -> None:
        pass

    def _write_channel(self, cmd: str) -> None:
        self._channel.write_channel(cmd)

    def _write_channel_floats(self, cmd: str, values: Sequence[float]) -> None:
        self._channel.write_channel_floats(cmd, values)

    def _ask_channel(self, cmd: str) -> str:
        return self._channel.ask_channel(cmd)

    def _channel_message(self, template: str) -> None:
        return self._channel._channel_message(template)


class _Dc_Context(_Channel_Context):

    def __init__(self, channel: "QDac2Channel"):
        super().__init__(channel)
        self._write_channel("sour{0}:dc:trig:sour hold")
        self._trigger: Optional[QDac2Trigger_Context] = None
        self._marker_start: Optional[QDac2Trigger_Context] = None
        self._marker_end: Optional[QDac2Trigger_Context] = None
        self._marker_step_start: Optional[QDac2Trigger_Context] = None
        self._marker_step_end: Optional[QDac2Trigger_Context] = None

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to DC generator

        Args:
            trigger (QDac2Trigger_Context): trigger that will start DC
        """
        self._trigger = trigger
        internal = _trigger_context_to_value(trigger)
        self._write_channel(f'sour{"{0}"}:dc:trig:sour int{internal}')
        self._make_ready_to_start()

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to DC generator

        Args:
            trigger (ExternalInput): trigger that will start DC generator
        """
        self._trigger = None
        self._write_channel(f'sour{"{0}"}:dc:trig:sour ext{trigger}')
        self._make_ready_to_start()

    def abort(self) -> None:
        """Abort any DC running generator on the channel"""
        self._write_channel("sour{0}:dc:abor")

    def end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of the DC generator

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end
        """
        if not self._marker_end:
            self._marker_end = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:dc:mark:end {self._marker_end.value}')
        return self._marker_end

    def start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of the DC generator

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning
        """
        if not self._marker_start:
            self._marker_start = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:dc:mark:star {self._marker_start.value}')
        return self._marker_start

    def step_end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of each step

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each step
        """
        if not self._marker_step_end:
            self._marker_step_end = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:dc:mark:send {self._marker_step_end.value}')
        return self._marker_step_end

    def step_start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of each step

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each step
        """
        if not self._marker_step_start:
            self._marker_step_start = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:dc:mark:sst {self._marker_step_start.value}')
        return self._marker_step_start

    def _set_delay(self, delay_s: float) -> None:
        self._write_channel(f'sour{"{0}"}:dc:del {delay_s}')

    def _set_triggering(self) -> None:
        self._write_channel("sour{0}:dc:trig:sour bus")
        self._make_ready_to_start()

    def _start(self, description: str) -> None:
        if self._trigger:
            self._make_ready_to_start()
            return self._write_channel(f"tint {self._trigger.value}")
        self._switch_to_immediate_trigger()
        self._write_channel("sour{0}:dc:init")

    def _make_ready_to_start(self) -> None:
        self._write_channel("sour{0}:dc:init:cont on")
        self._write_channel("sour{0}:dc:init")

    def _switch_to_immediate_trigger(self) -> None:
        self._write_channel("sour{0}:dc:init:cont off")
        self._write_channel("sour{0}:dc:trig:sour imm")


class Sweep_Context(_Dc_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        start_V: float,
        stop_V: float,
        points: int,
        repetitions: int,
        dwell_s: float,
        delay_s: float,
        backwards: bool,
        stepped: bool,
    ):
        self._repetitions = repetitions
        super().__init__(channel)
        channel.write_channel("sour{0}:volt:mode swe")
        self._set_voltages(start_V, stop_V)
        channel.write_channel(f'sour{"{0}"}:swe:poin {points}')
        self._set_trigger_mode(stepped)
        channel.write_channel(f'sour{"{0}"}:swe:dwel {dwell_s}')
        super()._set_delay(delay_s)
        self._set_direction(backwards)
        self._set_repetitions()
        self._set_triggering()

    def _set_voltages(self, start_V: float, stop_V: float):
        self._write_channel(f'sour{"{0}"}:swe:star {start_V}')
        self._write_channel(f'sour{"{0}"}:swe:stop {stop_V}')

    def _set_trigger_mode(self, stepped: bool) -> None:
        if stepped:
            return self._write_channel("sour{0}:swe:gen step")
        self._write_channel("sour{0}:swe:gen auto")

    def _set_direction(self, backwards: bool) -> None:
        if backwards:
            return self._write_channel("sour{0}:swe:dir down")
        self._write_channel("sour{0}:swe:dir up")

    def _set_repetitions(self) -> None:
        self._write_channel(f'sour{"{0}"}:swe:coun {self._repetitions}')

    def _perpetual(self) -> bool:
        return self._repetitions < 0

    def start(self) -> None:
        """Start the DC sweep"""
        self._start("DC sweep")

    def points(self) -> int:
        """
        Returns:
            int: Number of steps in the DC sweep
        """
        return int(self._ask_channel("sour{0}:swe:poin?"))

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the DC sweep
        """
        return int(self._ask_channel("sour{0}:swe:ncl?"))

    def time_s(self) -> float:
        """
        Returns:
            float: Seconds that it will take to do the sweep
        """
        return float(self._ask_channel("sour{0}:swe:time?"))

    def start_V(self) -> float:
        """
        Returns:
            float: Starting voltage
        """
        return float(self._ask_channel("sour{0}:swe:star?"))

    def stop_V(self) -> float:
        """
        Returns:
            float: Ending voltage
        """
        return float(self._ask_channel("sour{0}:swe:stop?"))

    def values_V(self) -> Sequence[float]:
        """
        Returns:
            Sequence[float]: List of voltages
        """
        return list(np.linspace(self.start_V(), self.stop_V(), self.points()))


class List_Context(_Dc_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        voltages: Sequence[float],
        repetitions: int,
        dwell_s: float,
        delay_s: float,
        backwards: bool,
        stepped: bool,
    ):
        super().__init__(channel)
        self._repetitions = repetitions
        self._write_channel("sour{0}:volt:mode list")
        self._set_voltages(voltages)
        self._set_trigger_mode(stepped)
        self._write_channel(f'sour{"{0}"}:list:dwel {dwell_s}')
        super()._set_delay(delay_s)
        self._set_direction(backwards)
        self._set_repetitions()
        self._set_triggering()

    def _set_voltages(self, voltages: Sequence[float]) -> None:
        self._write_channel_floats("sour{0}:list:volt ", voltages)

    def _set_trigger_mode(self, stepped: bool) -> None:
        if stepped:
            return self._write_channel("sour{0}:list:tmod step")
        self._write_channel("sour{0}:list:tmod auto")

    def _set_direction(self, backwards: bool) -> None:
        if backwards:
            return self._write_channel("sour{0}:list:dir down")
        self._write_channel("sour{0}:list:dir up")

    def _set_repetitions(self) -> None:
        self._write_channel(f'sour{"{0}"}:list:coun {self._repetitions}')

    def _perpetual(self) -> bool:
        return self._repetitions < 0

    def start(self) -> None:
        """Start the DC list generator"""
        self._start("DC list")

    def append(self, voltages: Sequence[float]) -> None:
        """Append voltages to the existing list

        Arguments:
            voltages (Sequence[float]): Sequence of voltages
        """
        self._write_channel_floats("sour{0}:list:volt:app ", voltages)
        self._make_ready_to_start()

    def points(self) -> int:
        """
        Returns:
            int: Number of steps in the DC list
        """
        return int(self._ask_channel("sour{0}:list:poin?"))

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the DC list
        """
        return int(self._ask_channel("sour{0}:list:ncl?"))

    def values_V(self) -> Sequence[float]:
        """
        Returns:
            Sequence[float]: List of voltages
        """
        # return comma_sequence_to_list_of_floats(
        #     self._ask_channel('sour{0}:list:volt?'))
        return comma_sequence_to_list_of_floats(self._ask_channel("sour{0}:list:volt?"))


class _Waveform_Context(_Channel_Context):

    def __init__(self, channel: "QDac2Channel"):
        super().__init__(channel)
        self._trigger: Optional[QDac2Trigger_Context] = None
        self._marker_start: Optional[QDac2Trigger_Context] = None
        self._marker_end: Optional[QDac2Trigger_Context] = None
        self._marker_period_start: Optional[QDac2Trigger_Context] = None
        self._marker_period_end: Optional[QDac2Trigger_Context] = None

    def _start(self, wave_kind: str, description: str) -> None:
        if self._trigger:
            self._make_ready_to_start(wave_kind)
            return self._write_channel(f"tint {self._trigger.value}")
        self._switch_to_immediate_trigger(wave_kind)
        self._write_channel(f'sour{"{0}"}:{wave_kind}:init')

    def _start_on(self, trigger: QDac2Trigger_Context, wave_kind: str) -> None:
        self._trigger = trigger
        internal = _trigger_context_to_value(trigger)
        self._write_channel(f'sour{"{0}"}:{wave_kind}:trig:sour int{internal}')
        self._make_ready_to_start(wave_kind)

    def _start_on_external(self, trigger: ExternalInput, wave_kind: str) -> None:
        self._trigger = None
        self._write_channel(f'sour{"{0}"}:{wave_kind}:trig:sour ext{trigger}')
        self._make_ready_to_start(wave_kind)

    def _end_marker(self, wave_kind: str) -> QDac2Trigger_Context:
        if not self._marker_end:
            self._marker_end = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:{wave_kind}:mark:end {self._marker_end.value}')
        return self._marker_end

    def _start_marker(self, wave_kind: str) -> QDac2Trigger_Context:
        if not self._marker_start:
            self._marker_start = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:{wave_kind}:mark:star {self._marker_start.value}')
        return self._marker_start

    def _period_end_marker(self, wave_kind: str) -> QDac2Trigger_Context:
        if not self._marker_period_end:
            self._marker_period_end = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:{wave_kind}:mark:pend {self._marker_period_end.value}')
        return self._marker_period_end

    def _period_start_marker(self, wave_kind: str) -> QDac2Trigger_Context:
        if not self._marker_period_start:
            self._marker_period_start = self.allocate_trigger()
        self._write_channel(f'sour{"{0}"}:{wave_kind}:mark:pstart {self._marker_period_start.value}')
        return self._marker_period_start

    def _make_ready_to_start(self, wave_kind: str) -> None:
        self._write_channel(f'sour{"{0}"}:{wave_kind}:init:cont on')
        self._write_channel(f'sour{"{0}"}:{wave_kind}:init')

    def _switch_to_immediate_trigger(self, wave_kind: str):
        self._write_channel(f'sour{"{0}"}:{wave_kind}:init:cont off')
        self._write_channel(f'sour{"{0}"}:{wave_kind}:trig:sour imm')

    def _set_delay(self, wave_kind: str, delay_s) -> None:
        self._write_channel(f'sour{"{0}"}:{wave_kind}:del {delay_s}')

    def _set_slew(self, wave_kind: str, slew_V_s: Optional[float]) -> None:
        if slew_V_s:
            # Bug, see https://trello.com/c/SeeUrRNY
            self._write_channel(f'sour{"{0}"}:{wave_kind}:slew {slew_V_s}')
        else:
            self._write_channel(f'sour{"{0}"}:{wave_kind}:slew inf')


class Square_Context(_Waveform_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        frequency_Hz: Optional[float],
        repetitions: int,
        period_s: Optional[float],
        duty_cycle_percent: float,
        kind: str,
        inverted: bool,
        span_V: float,
        offset_V: float,
        delay_s: float,
        slew_V_s: Optional[float],
    ):
        super().__init__(channel)
        self._repetitions = repetitions
        self._write_channel("sour{0}:squ:trig:sour hold")
        self._set_frequency(frequency_Hz, period_s)
        self._write_channel(f'sour{"{0}"}:squ:dcyc {duty_cycle_percent}')
        self._set_type(kind)
        self._set_polarity(inverted)
        self._write_channel(f'sour{"{0}"}:squ:span {span_V}')
        self._write_channel(f'sour{"{0}"}:squ:offs {offset_V}')
        self._set_slew("squ", slew_V_s)
        super()._set_delay("squ", delay_s)
        self._write_channel(f'sour{"{0}"}:squ:coun {repetitions}')
        self._set_triggering()

    def start(self) -> None:
        """Start the square wave generator"""
        self._start("squ", "square wave")

    def abort(self) -> None:
        """Abort any running square wave generator"""
        self._write_channel("sour{0}:squ:abor")

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the square wave
        """
        return int(self._ask_channel("sour{0}:squ:ncl?"))

    def _set_frequency(self, frequency_Hz: Optional[float], period_s: Optional[float]) -> None:
        if frequency_Hz:
            return self._write_channel(f'sour{"{0}"}:squ:freq {frequency_Hz}')
        if period_s:
            self._write_channel(f'sour{"{0}"}:squ:per {period_s}')

    def _set_type(self, kind: str) -> None:
        if kind == "positive":
            self._write_channel("sour{0}:squ:typ pos")
        elif kind == "negative":
            self._write_channel("sour{0}:squ:typ neg")
        else:
            self._write_channel("sour{0}:squ:typ symm")

    def _set_polarity(self, inverted: bool) -> None:
        if inverted:
            self._write_channel("sour{0}:squ:pol inv")
        else:
            self._write_channel("sour{0}:squ:pol norm")

    def _set_triggering(self) -> None:
        self._write_channel("sour{0}:squ:trig:sour bus")
        self._make_ready_to_start("squ")

    def end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of the square wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end
        """
        return super()._end_marker("squ")

    def start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of the square wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning
        """
        return super()._start_marker("squ")

    def period_end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each period
        """
        return super()._period_end_marker("squ")

    def period_start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning of each period
        """
        return super()._period_start_marker("squ")

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to start the square wave generator

        Args:
            trigger (QDac2Trigger_Context): trigger that will start square wave
        """
        return super()._start_on(trigger, "squ")

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to start the square wave generator

        Args:
            trigger (ExternalInput): external trigger that will start square wave
        """
        return super()._start_on_external(trigger, "squ")


class Sine_Context(_Waveform_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        frequency_Hz: Optional[float],
        repetitions: int,
        period_s: Optional[float],
        inverted: bool,
        span_V: float,
        offset_V: float,
        delay_s: float,
        slew_V_s: Optional[float],
    ):
        super().__init__(channel)
        self._repetitions = repetitions
        self._write_channel("sour{0}:sine:trig:sour hold")
        self._set_frequency(frequency_Hz, period_s)
        self._set_polarity(inverted)
        self._write_channel(f'sour{"{0}"}:sine:span {span_V}')
        self._write_channel(f'sour{"{0}"}:sine:offs {offset_V}')
        self._set_slew("sine", slew_V_s)
        super()._set_delay("sine", delay_s)
        self._write_channel(f'sour{"{0}"}:sine:coun {repetitions}')
        self._set_triggering()

    def start(self) -> None:
        """Start the sine wave generator"""
        self._start("sine", "sine wave")

    def abort(self) -> None:
        """Abort any running sine wave generator"""
        self._write_channel("sour{0}:sine:abor")

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the sine wave
        """
        return int(self._ask_channel("sour{0}:sine:ncl?"))

    def _set_frequency(self, frequency_Hz: Optional[float], period_s: Optional[float]) -> None:
        if frequency_Hz:
            return self._write_channel(f'sour{"{0}"}:sine:freq {frequency_Hz}')
        if period_s:
            self._write_channel(f'sour{"{0}"}:sine:per {period_s}')

    def _set_polarity(self, inverted: bool) -> None:
        if inverted:
            self._write_channel("sour{0}:sine:pol inv")
        else:
            self._write_channel("sour{0}:sine:pol norm")

    def _set_triggering(self) -> None:
        self._write_channel("sour{0}:sine:trig:sour bus")
        self._make_ready_to_start("sine")

    def end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of the sine wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end
        """
        return super()._end_marker("sine")

    def start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of the sine wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning
        """
        return super()._start_marker("sine")

    def period_end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each period
        """
        return super()._period_end_marker("sine")

    def period_start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning of each period
        """
        return super()._period_start_marker("sine")

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to start the sine wave generator

        Args:
            trigger (QDac2Trigger_Context): trigger that will start sine wave
        """
        return super()._start_on(trigger, "sine")

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to start the sine wave generator

        Args:
            trigger (ExternalInput): external trigger that will start sine wave
        """
        return super()._start_on_external(trigger, "sine")


class Triangle_Context(_Waveform_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        frequency_Hz: Optional[float],
        repetitions: int,
        period_s: Optional[float],
        duty_cycle_percent: float,
        inverted: bool,
        span_V: float,
        offset_V: float,
        delay_s: float,
        slew_V_s: Optional[float],
    ):
        super().__init__(channel)
        self._repetitions = repetitions
        self._write_channel("sour{0}:tri:trig:sour hold")
        self._set_frequency(frequency_Hz, period_s)
        self._write_channel(f'sour{"{0}"}:tri:dcyc {duty_cycle_percent}')
        self._set_polarity(inverted)
        self._write_channel(f'sour{"{0}"}:tri:span {span_V}')
        self._write_channel(f'sour{"{0}"}:tri:offs {offset_V}')
        self._set_slew("tri", slew_V_s)
        super()._set_delay("tri", delay_s)
        self._write_channel(f'sour{"{0}"}:tri:coun {repetitions}')
        self._set_triggering()

    def start(self) -> None:
        """Start the triangle wave generator"""
        self._start("tri", "triangle wave")

    def abort(self) -> None:
        """Abort any running triangle wave generator"""
        self._write_channel("sour{0}:tri:abor")

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the triangle wave
        """
        return int(self._ask_channel("sour{0}:tri:ncl?"))

    def _set_frequency(self, frequency_Hz: Optional[float], period_s: Optional[float]) -> None:
        if frequency_Hz:
            return self._write_channel(f'sour{"{0}"}:tri:freq {frequency_Hz}')
        if period_s:
            self._write_channel(f'sour{"{0}"}:tri:per {period_s}')

    def _set_type(self, kind: bool) -> None:
        if kind == "positive":
            self._write_channel("sour{0}:tri:typ pos")
        elif kind == "negative":
            self._write_channel("sour{0}:tri:typ neg")
        else:
            self._write_channel("sour{0}:tri:typ symm")

    def _set_polarity(self, inverted: bool) -> None:
        if inverted:
            self._write_channel("sour{0}:tri:pol inv")
        else:
            self._write_channel("sour{0}:tri:pol norm")

    def _set_triggering(self) -> None:
        self._write_channel("sour{0}:tri:trig:sour bus")
        self._make_ready_to_start("tri")

    def end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of the triangle wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end
        """
        return super()._end_marker("tri")

    def start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of the triangle wave

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning
        """
        return super()._start_marker("tri")

    def period_end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each period
        """
        return super()._period_end_marker("tri")

    def period_start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning of each period
        """
        return super()._period_start_marker("tri")

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to start the triangle wave generator

        Args:
            trigger (QDac2Trigger_Context): trigger that will start triangle
        """
        return super()._start_on(trigger, "tri")

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to start the triangle wave generator

        Args:
            trigger (ExternalInput): external trigger that will start triangle
        """
        return super()._start_on_external(trigger, "tri")


class Awg_Context(_Waveform_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        trace_name: str,
        repetitions: int,
        scale: float,
        offset_V: float,
        slew_V_s: Optional[float],
    ):
        super().__init__(channel)
        self._repetitions = repetitions
        self._write_channel("sour{0}:awg:trig:sour hold")
        self._write_channel(f'sour{"{0}"}:awg:def "{trace_name}"')
        self._write_channel(f'sour{"{0}"}:awg:scal {scale}')
        self._write_channel(f'sour{"{0}"}:awg:offs {offset_V}')
        self._set_slew("awg", slew_V_s)
        self._write_channel(f'sour{"{0}"}:awg:coun {repetitions}')
        self._set_triggering()

    def start(self) -> None:
        """Start the AWG"""
        self._start("awg", "AWG")

    def abort(self) -> None:
        """Abort any running AWG"""
        self._write_channel("sour{0}:awg:abor")

    def cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of cycles remaining in the AWG
        """
        return int(self._ask_channel("sour{0}:awg:ncl?"))

    def _set_triggering(self) -> None:
        self._write_channel("sour{0}:awg:trig:sour bus")
        self._make_ready_to_start("awg")

    def end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of the AWG

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end
        """
        return super()._end_marker("awg")

    def start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of the AWG

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning
        """
        return super()._start_marker("awg")

    def period_end_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the end of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the end of each period
        """
        return super()._period_end_marker("awg")

    def period_start_marker(self) -> QDac2Trigger_Context:
        """Internal trigger that will mark the beginning of each period

        A new internal trigger is allocated if necessary.

        Returns:
            QDac2Trigger_Context: trigger that will mark the beginning of each period
        """
        return super()._period_start_marker("awg")

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to start the AWG

        Args:
            trigger (QDac2Trigger_Context): trigger that will start AWG
        """
        return super()._start_on(trigger, "awg")

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to start the AWG

        Args:
            trigger (ExternalInput): external trigger that will start AWG
        """
        return super()._start_on_external(trigger, "awg")


class Measurement_Context(_Channel_Context):

    def __init__(
        self,
        channel: "QDac2Channel",
        delay_s: float,
        repetitions: int,
        current_range: str,
        aperture_s: Optional[float],
        nplc: Optional[int],
    ):
        super().__init__(channel)
        self._trigger: Optional[QDac2Trigger_Context] = None
        self._write_channel(f'sens{"{0}"}:del {delay_s}')
        self._write_channel(f'sens{"{0}"}:rang {current_range}')
        self._set_aperture(aperture_s, nplc)
        self._write_channel(f'sens{"{0}"}:coun {repetitions}')
        self._set_triggering()

    def start(self) -> None:
        """Start a current measurement"""
        if self._trigger:
            return self._write_channel(f"tint {self._trigger.value}")
        self._switch_to_immediate_trigger()
        self._write_channel("sens{0}:init")

    def _switch_to_immediate_trigger(self) -> None:
        self._write_channel("sens{0}:init:cont off")
        self._write_channel("sens{0}:trig:sour imm")

    def start_on(self, trigger: QDac2Trigger_Context) -> None:
        """Attach internal trigger to start the current measurement

        Args:
            trigger (QDac2Trigger_Context): trigger that will start measurement
        """
        self._trigger = trigger
        internal = _trigger_context_to_value(trigger)
        self._write_channel(f'sens{"{0}"}:trig:sour int{internal}')
        self._write_channel(f'sens{"{0}"}:init:cont on')
        self._write_channel(f'sens{"{0}"}:init')

    def start_on_external(self, trigger: ExternalInput) -> None:
        """Attach external trigger to start the current measurement

        Args:
            trigger (ExternalInput): trigger that will start measurement
        """
        self._write_channel(f'sens{"{0}"}:trig:sour ext{trigger}')
        self._write_channel(f'sens{"{0}"}:init:cont on')
        self._write_channel(f'sens{"{0}"}:init')

    def abort(self) -> None:
        """Abort current measurement"""
        self._write_channel("sens{0}:abor")

    def n_cycles_remaining(self) -> int:
        """
        Returns:
            int: Number of measurements remaining
        """
        return int(self._ask_channel("sens{0}:ncl?"))

    def n_available(self) -> int:
        """
        Returns:
            int: Number of measurements available
        """
        return int(self._ask_channel("sens{0}:data:poin?"))

    def available_A(self) -> Sequence[float]:
        """Retrieve current measurements

        The available measurements will be removed from measurement queue.

        Returns:
            Sequence[float]: list of available current measurements
        """
        # Bug circumvention
        if self.n_available() == 0:
            return list()
        return comma_sequence_to_list_of_floats(self._ask_channel("sens{0}:data:rem?"))

    def peek_A(self) -> float:
        """Peek at the first available current measurement

        Returns:
            float: current in Amperes
        """
        return float(self._ask_channel("sens{0}:data:last?"))

    def _set_aperture(self, aperture_s: Optional[float], nplc: Optional[int]) -> None:
        if aperture_s:
            return self._write_channel(f'sens{"{0}"}:aper {aperture_s}')
        self._write_channel(f'sens{"{0}"}:nplc {nplc}')

    def _set_triggering(self) -> None:
        self._write_channel("sens{0}:trig:sour bus")
        self._write_channel("sens{0}:init")


class QDac2Channel(InstrumentChannel):

    def __init__(self, parent: "QDac2", name: str, channum: int):
        super().__init__(parent, name)
        self._channum = channum
        self.add_parameter(
            name="measurement_range",
            label="range",
            set_cmd="sens{1}:rang {0}".format("{}", channum),
            get_cmd=f"sens{channum}:rang?",
            vals=validators.Enum("low", "high"),
        )
        self.add_parameter(
            name="measurement_aperture_s",
            label="aperture",
            unit="s",
            set_cmd="sens{1}:aper {0}".format("{}", channum),
            get_cmd=f"sens{channum}:aper?",
            get_parser=float,
        )
        self.add_parameter(
            name="measurement_nplc",
            label="PLC",
            set_cmd="sens{1}:nplc {0}".format("{}", channum),
            get_cmd=f"sens{channum}:nplc?",
            get_parser=int,
        )
        self.add_parameter(
            name="measurement_delay_s",
            label="delay",
            unit="s",
            set_cmd="sens{1}:del {0}".format("{}", channum),
            get_cmd=f"sens{channum}:del?",
            get_parser=float,
        )
        self.add_function(name="measurement_abort", call_cmd=f"sens{channum}:abor")
        self.add_parameter(
            name="measurement_count",
            label="count",
            set_cmd="sens{1}:coun {0}".format("{}", channum),
            get_cmd=f"sens{channum}:coun?",
            get_parser=int,
        )
        self.add_parameter(
            name="n_masurements_remaining", label="remaning", get_cmd=f"sens{channum}:ncl?", get_parser=int
        )
        self.add_parameter(
            name="current_last_A", label="last", unit="A", get_cmd=f"sens{channum}:data:last?", get_parser=float
        )
        self.add_parameter(
            name="n_measurements_available", label="available", get_cmd=f"sens{channum}:data:poin?", get_parser=int
        )
        self.add_parameter(
            name="current_start_on",
            # Channel {channum} current measurement on internal trigger
            set_parser=_trigger_context_to_value,
            set_cmd="sens{1}:trig:sour int{0}".format("{}", channum),
        )
        self.add_parameter(
            name="measurement_start_on_external",
            # Channel {channum} current measurement on external input
            set_cmd="sens{1}:trig:sour ext{0}".format("{}", channum),
        )
        self.add_parameter(
            name="output_range",
            label="range",
            set_cmd="sour{1}:rang {0}".format("{}", channum),
            get_cmd=f"sour{channum}:rang?",
            vals=validators.Enum("low", "high"),
        )
        self.add_parameter(
            name="output_low_range_minimum_V",
            label="low range min",
            unit="V",
            get_cmd=f"sour{channum}:rang:low:min?",
            get_parser=float,
        )
        self.add_parameter(
            name="output_low_range_maximum_V",
            label="low voltage max",
            unit="V",
            get_cmd=f"sour{channum}:rang:low:max?",
            get_parser=float,
        )
        self.add_parameter(
            name="output_high_range_minimum_V",
            label="high voltage min",
            unit="V",
            get_cmd=f"sour{channum}:rang:high:min?",
            get_parser=float,
        )
        self.add_parameter(
            name="output_high_range_maximum_V",
            label="high voltage max",
            unit="V",
            get_cmd=f"sour{channum}:rang:high:max?",
            get_parser=float,
        )
        self.add_parameter(
            name="output_filter",
            label="low-pass cut-off",
            unit="Hz",
            set_cmd="sour{1}:filt {0}".format("{}", channum),
            get_cmd=f"sour{channum}:filt?",
            get_parser=str,
            vals=validators.Enum("dc", "med", "high"),
        )
        self.add_parameter(
            name="dc_constant_V",
            label=f"ch{channum}",
            unit="V",
            set_cmd=self._set_fixed_voltage_immediately,
            get_cmd=f"sour{channum}:volt?",
            get_parser=float,
            vals=validators.Numbers(-10.0, 10.0),
        )
        self.add_parameter(
            name="dc_last_V", label=f"ch{channum}", unit="V", get_cmd=f"sour{channum}:volt:last?", get_parser=float
        )
        self.add_parameter(
            name="dc_next_V",
            label=f"ch{channum}",
            unit="V",
            set_cmd="sour{1}:volt:trig {0}".format("{}", channum),
            get_cmd=f"sour{channum}:volt:trig?",
            get_parser=float,
        )
        self.add_parameter(
            name="dc_slew_rate_V_per_s",
            label=f"ch{channum}",
            unit="V/s",
            set_cmd="sour{1}:volt:slew {0}".format("{}", channum),
            get_cmd=f"sour{channum}:volt:slew?",
            get_parser=float,
        )
        self.add_parameter(
            name="read_current_A",
            # Perform immediate current measurement on channel
            label=f"ch{channum}",
            unit="A",
            get_cmd=f"read{channum}?",
            get_parser=float,
        )
        self.add_parameter(
            name="fetch_current_A",
            # Retrieve all available current measurements on channel
            label=f"ch{channum}",
            unit="A",
            get_cmd=f"fetc{channum}?",
            get_parser=comma_sequence_to_list_of_floats,
        )
        self.add_parameter(
            name="dc_mode",
            label="DC mode",
            set_cmd="sour{1}:volt:mode {0}".format("{}", channum),
            get_cmd=f"sour{channum}:volt:mode?",
            vals=validators.Enum("fixed", "list", "sweep"),
        )
        self.add_function(name="dc_initiate", call_cmd=f"sour{channum}:dc:init")
        self.add_function(name="dc_abort", call_cmd=f"sour{channum}:dc:abor")
        self.add_function(name="abort", call_cmd=f"sour{channum}:all:abor")

    @property
    def number(self) -> int:
        """Channel number"""
        return self._channum

    def clear_measurements(self) -> Sequence[float]:
        """Retrieve current measurements

        The available measurements will be removed from measurement queue.

        Returns:
            Sequence[float]: list of available current measurements
        """
        # Bug circumvention
        if int(self.ask_channel("sens{0}:data:poin?")) == 0:
            return list()
        return comma_sequence_to_list_of_floats(self.ask_channel("sens{0}:data:rem?"))

    def measurement(
        self,
        delay_s: float = 0.0,
        repetitions: int = 1,
        current_range: str = "high",
        aperture_s: Optional[float] = None,
        nplc: Optional[int] = None,
    ) -> Measurement_Context:
        """Set up a sequence of current measurements

        Args:
            delay_s (float, optional): Seconds to delay the actual measurement after trigger (default 0)
            repetitions (int, optional): Number of consecutive measurements (default 1)
            current_range (str, optional): high (10mA, default) or low (200nA)
            nplc (None, optional): Integration time in power-line cycles (default 1)
            aperture_s (None, optional): Seconds of integration time instead of NPLC

        Returns:
            Measurement_Context: context manager

        Raises:
            ValueError: configuration error
        """
        if aperture_s and nplc:
            raise ValueError("Only one of nplc or aperture_s can be " "specified for a current measurement")
        if not aperture_s and not nplc:
            nplc = 1
        return Measurement_Context(self, delay_s, repetitions, current_range, aperture_s, nplc)

    def output_mode(self, range: str = "high", filter: str = "high") -> None:
        """Set the output voltage

        Args:
            range (str, optional): Low or high (default) current range
            filter (str, optional): DC (10Hz), medium (10kHz) or high (300kHz, default) voltage filter
        """
        self.output_range(range)
        self.output_filter(filter)

    def dc_list(
        self,
        voltages: Sequence[float],
        repetitions: int = 1,
        dwell_s: float = 1e-03,
        delay_s: float = 0,
        backwards: bool = False,
        stepped: bool = False,
    ) -> List_Context:
        """Set up a DC-list generator

        Args:
            voltages (Sequence[float]): Voltages in list
            repetitions (int, optional): Number of repetitions of the list (default 1)
            dwell_s (float, optional): Seconds between each voltage (default 1ms)
            delay_s (float, optional): Seconds of delay after receiving a trigger (default 0)
            backwards (bool, optional): Use list in reverse (default is forward)
            stepped (bool, optional): True means that each step needs to be triggered (default False)

        Returns:
            List_Context: context manager
        """
        return List_Context(self, voltages, repetitions, dwell_s, delay_s, backwards, stepped)

    def dc_sweep(
        self,
        start_V: float,
        stop_V: float,
        points: int,
        repetitions: int = 1,
        dwell_s: float = 1e-03,
        delay_s: float = 0,
        backwards=False,
        stepped=True,
    ) -> Sweep_Context:
        """Set up a DC sweep

        Args:
            start_V (float): Start voltage
            stop_V (float): Send voltage
            points (int): Number of steps
            repetitions (int, optional): Number of repetition (default 1)
            dwell_s (float, optional): Seconds between each voltage (default 1ms)
            delay_s (float, optional): Seconds of delay after receiving a trigger (default 0)
            backwards (bool, optional): Sweep in reverse (default is forward)
            stepped (bool, optional): True means that each step needs to be triggered (default False)

        Returns:
            Sweep_Context: context manager
        """
        return Sweep_Context(self, start_V, stop_V, points, repetitions, dwell_s, delay_s, backwards, stepped)

    def square_wave(
        self,
        frequency_Hz: Optional[float] = None,
        period_s: Optional[float] = None,
        repetitions: int = -1,
        duty_cycle_percent: float = 50.0,
        kind: str = "symmetric",
        inverted: bool = False,
        span_V: float = 0.2,
        offset_V: float = 0.0,
        delay_s: float = 0,
        slew_V_s: Optional[float] = None,
    ) -> Square_Context:
        """Set up a square-wave generator

        Args:
            frequency_Hz (float, optional): Frequency
            period_s (float, optional): Period length (instead of frequency)
            repetitions (int, optional): Number of repetition (default infinite)
            duty_cycle_percent (float, optional): Percentage on-time (default 50%)
            kind (str, optional): Positive, negative or symmetric (default) around the offset
            inverted (bool, optional): True means flipped (default False)
            span_V (float, optional): Voltage span (default 200mV)
            offset_V (float, optional): Offset (default 0V)
            delay_s (float, optional): Seconds of delay after receiving a trigger (default 0)
            slew_V_s (float, optional): Max slew rate in V/s (default None)

        Returns:
            Square_Context: context manager

        Raises:
            ValueError: configuration error
        """
        if frequency_Hz and period_s:
            raise ValueError(error_ambiguous_wave)
        if not frequency_Hz and not period_s:
            frequency_Hz = 1000
        return Square_Context(
            self,
            frequency_Hz,
            repetitions,
            period_s,
            duty_cycle_percent,
            kind,
            inverted,
            span_V,
            offset_V,
            delay_s,
            slew_V_s,
        )

    def sine_wave(
        self,
        frequency_Hz: Optional[float] = None,
        period_s: Optional[float] = None,
        repetitions: int = -1,
        inverted: bool = False,
        span_V: float = 0.2,
        offset_V: float = 0.0,
        delay_s: float = 0,
        slew_V_s: Optional[float] = None,
    ) -> Sine_Context:
        """Set up a sine-wave generator

        Args:
            frequency_Hz (float, optional): Frequency
            period_s (float, optional): Period length (instead of frequency)
            repetitions (int, optional): Number of repetition (default infinite)
            inverted (bool, optional): True means flipped (default False)
            span_V (float, optional): Voltage span (default 200mV)
            offset_V (float, optional): Offset (default 0V)
            delay_s (float, optional): Seconds of delay after receiving a trigger (default 0)
            slew_V_s (None, optional): Max slew rate in V/s (default None)

        Returns:
            Sine_Context: context manager

        Raises:
            ValueError: configuration error
        """
        if frequency_Hz and period_s:
            raise ValueError(error_ambiguous_wave)
        if not frequency_Hz and not period_s:
            frequency_Hz = 1000
        return Sine_Context(self, frequency_Hz, repetitions, period_s, inverted, span_V, offset_V, delay_s, slew_V_s)

    def triangle_wave(
        self,
        frequency_Hz: Optional[float] = None,
        period_s: Optional[float] = None,
        repetitions: int = -1,
        duty_cycle_percent: float = 50.0,
        inverted: bool = False,
        span_V: float = 0.2,
        offset_V: float = 0.0,
        delay_s: float = 0,
        slew_V_s: Optional[float] = None,
    ) -> Triangle_Context:
        """Set up a triangle-wave generator

        Args:
            frequency_Hz (float, optional): Frequency
            period_s (float, optional): Period length (instead of frequency)
            repetitions (int, optional): Number of repetition (default infinite)
            duty_cycle_percent (float, optional): Percentage on-time (default 50%)
            inverted (bool, optional): True means flipped (default False)
            span_V (float, optional): Voltage span (default 200mV)
            offset_V (float, optional): Offset (default 0V)
            delay_s (float, optional): Seconds of delay after receiving a trigger (default 0)
            slew_V_s (float, optional): Max slew rate in V/s (default None)

        Returns:
            Triangle_Context: context manager

        Raises:
            ValueError: configuration error
        """
        if frequency_Hz and period_s:
            raise ValueError(error_ambiguous_wave)
        if not frequency_Hz and not period_s:
            frequency_Hz = 1000
        return Triangle_Context(
            self, frequency_Hz, repetitions, period_s, duty_cycle_percent, inverted, span_V, offset_V, delay_s, slew_V_s
        )

    def arbitrary_wave(
        self,
        trace_name: str,
        repetitions: int = 1,
        scale: float = 1.0,
        offset_V: float = 0.0,
        slew_V_s: Optional[float] = None,
    ) -> Awg_Context:
        """Set up an arbitrary-wave generator

        Args:
            trace_name (str): Use data from this named trace
            repetitions (int, optional): Number of repetition (default 1)
            scale (float, optional): Scaling factor of voltages (default 1)
            offset_V (float, optional): Offset (default 0V)
            slew_V_s (None, optional): Max slew rate in V/s (default None)

        Returns:
            Awg_Context: context manager
        """
        return Awg_Context(self, trace_name, repetitions, scale, offset_V, slew_V_s)

    def _set_fixed_voltage_immediately(self, v) -> None:
        self.write(f"sour{self._channum}:volt:mode fix")
        self.write(f"sour{self._channum}:volt {v}")

    def ask_channel(self, cmd: str) -> str:
        """Inject channel number into SCPI query

        Arguments:
            cmd (str): Must contain a '{0}' placeholder for the channel number

        Returns:
            str: SCPI answer
        """
        return self.ask(self._channel_message(cmd))

    def write_channel(self, cmd: str) -> None:
        """Inject channel number into SCPI command

        Arguments:
            cmd (str): Must contain a '{0}' placeholder for the channel number
        """
        self.write(self._channel_message(cmd))

    def write_channel_floats(self, cmd: str, values: Sequence[float]) -> None:
        """Inject channel number and a list of values into SCPI command

        The values are appended to the end of the command.

        Arguments:
            cmd (str): Must contain a '{0}' placeholder for channel number
            values (Sequence[float]): Sequence of numbers
        """
        self._parent.write_floats(self._channel_message(cmd), values)

    def write(self, cmd: str) -> None:
        """Send a SCPI command

        Args:
            cmd (str): SCPI command
        """
        self._parent.write(cmd)

    def _channel_message(self, template: str):
        return template.format(self._channum)


class Trace_Context:

    def __init__(self, parent, name: str, size: int):
        self._parent = parent
        self._size = size
        self._name = name
        self._parent.write(f'trac:def "{name}",{size}')

    def __len__(self):
        return self.size

    @property
    def size(self) -> int:
        """Number of values in trace"""
        return self._size

    @property
    def name(self) -> str:
        """Name of trace"""
        return self._name

    def waveform(self, values: Sequence[float]) -> None:
        """Fill values into trace

        Args:
            values (Sequence[float]): Sequence of values

        Raises:
            ValueError: size mismatch
        """
        if len(values) != self.size:
            raise ValueError(f"trace length {len(values)} does not match " f"allocated length {self.size}")
        self._parent.write_floats(f'trac:data "{self.name}",', values)


class Virtual_Sweep_Context:

    def __init__(
        self,
        arrangement: "Arrangement_Context",
        sweep: np.ndarray,
        start_trigger: Optional[str],
        step_time_s: float,
        step_trigger: Optional[str],
        repetitions: Optional[int],
    ):
        self._arrangement = arrangement
        self._sweep = sweep
        self._step_trigger = step_trigger
        self._step_time_s = step_time_s
        self._repetitions = repetitions
        self._allocate_triggers(start_trigger)
        self._qdac_ready = False

    def __enter__(self):
        self._ensure_qdac_setup()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Let Arrangement take care of freeing triggers
        return False

    def actual_values_V(self, contact: str) -> np.ndarray:
        """The corrected values that would actually be sent to the contact

        Args:
            contact (str): Name of contact

        Returns:
            np.ndarray: Corrected voltages
        """
        index = self._arrangement._contact_index(contact)
        return self._sweep[:, index]

    def start(self) -> None:
        """Start the 2D sweep"""
        self._ensure_qdac_setup()
        trigger = self._arrangement.get_trigger_by_name(self._start_trigger_name)
        self._arrangement._qdac.trigger(trigger)

    def _allocate_triggers(self, start_sweep: Optional[str]) -> None:
        if not start_sweep:
            # Use a random, unique name
            start_sweep = uuid.uuid4().hex
        self._arrangement._allocate_internal_triggers([start_sweep])
        self._start_trigger_name = start_sweep

    def _ensure_qdac_setup(self) -> None:
        if self._qdac_ready:
            return self._make_ready_to_start()
        self._route_inner_trigger()
        self._send_lists_to_qdac()
        self._qdac_ready = True

    def _route_inner_trigger(self) -> None:
        if not self._step_trigger:
            return
        trigger = self._arrangement.get_trigger_by_name(self._step_trigger)
        # All channels change in sync, so just use the first channel to make the
        # external trigger.
        channel = self._get_channel(0)
        channel.write_channel(f'sour{"{0}"}:dc:mark:sst ' f"{_trigger_context_to_value(trigger)}")

    def _get_channel(self, contact_index: int) -> "QDac2Channel":
        channel_number = self._arrangement._channels[contact_index]
        qdac = self._arrangement._qdac
        return qdac.channel(channel_number)

    def _send_lists_to_qdac(self) -> None:
        for contact_index in range(self._arrangement.shape):
            self._send_list_to_qdac(contact_index, self._sweep[:, contact_index])

    def _send_list_to_qdac(self, contact_index, voltages):
        channel = self._get_channel(contact_index)
        dc_list = channel.dc_list(voltages=voltages, dwell_s=self._step_time_s, repetitions=self._repetitions)
        trigger = self._arrangement.get_trigger_by_name(self._start_trigger_name)
        dc_list.start_on(trigger)

    def _make_ready_to_start(self):  # Bug circumvention
        for contact_index in range(self._arrangement.shape):
            channel = self._get_channel(contact_index)
            channel.write_channel("sour{0}:dc:init")


class Arrangement_Context:
    def __init__(
        self,
        qdac: "QDac2",
        contacts: dict[str, int],
        output_triggers: Optional[dict[str, int]],
        internal_triggers: Optional[Sequence[str]],
    ):
        self._qdac = qdac
        self._fix_contact_order(contacts)
        self._allocate_triggers(internal_triggers, output_triggers)
        self._correction = np.identity(self.shape)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._free_triggers()
        return False

    @property
    def shape(self) -> int:
        """Number of contacts in the arrangement"""
        return len(self._contacts)

    @property
    def correction_matrix(self) -> np.ndarray:
        """Correction matrix"""
        return self._correction

    @property
    def contact_names(self) -> Sequence[str]:
        """
        Returns:
            Sequence[str]: Contact names in the same order as channel_numbers
        """
        return self._contact_names

    def _allocate_internal_triggers(self, internal_triggers: Optional[Sequence[str]]) -> None:
        if not internal_triggers:
            return
        for name in internal_triggers:
            self._internal_triggers[name] = self._qdac.allocate_trigger()

    def initiate_correction(self, contact: str, factors: Sequence[float]) -> None:
        """Override how much a particular contact influences the other contacts

        Args:
            contact (str): Name of contact
            factors (Sequence[float]): factors between -1.0 and 1.0
        """
        index = self._contact_index(contact)
        self._correction[index] = factors

    def set_virtual_voltage(self, contact: str, voltage: float) -> None:
        """Set virtual voltage on specific contact

        The actual voltage that the contact will receive depends on the
        correction matrix.

        Args:
            contact (str): Name of contact
            voltage (float): Voltage corresponding to no correction
        """
        try:
            index = self._contact_index(contact)
        except KeyError:
            raise ValueError(f'No contact named "{contact}"')
        self._effectuate_virtual_voltage(index, voltage)

    def set_virtual_voltages(self, contacts_to_voltages: dict[str, float]) -> None:
        """Set virtual voltages on specific contacts in one go

        The actual voltage that each contact will receive depends on the
        correction matrix.

        Args:
            contact_to_voltages (Dict[str,float]): contact to voltage map
        """
        for contact, voltage in contacts_to_voltages.items():
            try:
                index = self._contact_index(contact)
            except KeyError:
                raise ValueError(f'No contact named "{contact}"')
            self._virtual_voltages[index] = voltage
        self._effectuate_virtual_voltages()

    def _effectuate_virtual_voltage(self, index: int, voltage: float) -> None:
        self._virtual_voltages[index] = voltage
        self._effectuate_virtual_voltages()

    def _effectuate_virtual_voltages(self) -> None:
        for index, channel_number in enumerate(self._channels):
            actual_V = self.actual_voltages()[index]
            self._qdac.channel(channel_number).dc_constant_V(actual_V)

    def add_correction(self, contact: str, factors: Sequence[float]) -> None:
        """Update how much a particular contact influences the other contacts

        This is mostly useful in arrangements where each contact has significant
        effect only on nearby contacts, and thus can be added incrementally.

        The factors are extended by the identity matrix and multiplied to the
        correction matrix.

        Args:
            contact (str): Name of contact
            factors (Sequence[float]): factors usually between -1.0 and 1.0
        """
        index = self._contact_index(contact)
        multiplier = np.identity(self.shape)
        multiplier[index] = factors
        self._correction = np.matmul(multiplier, self._correction)

    def _fix_contact_order(self, contacts: dict[str, int]) -> None:
        self._contact_names = list()
        self._contacts = dict()
        self._channels = list()
        index = 0
        for contact, channel in contacts.items():
            self._contact_names.append(contact)
            self._contacts[contact] = index
            index += 1
            self._channels.append(channel)
        self._virtual_voltages = np.zeros(self.shape)

    @property
    def channel_numbers(self) -> Sequence[int]:
        """
        Returns:
            Sequence[int]: Channels numbers in the same order as contact_names
        """
        return self._channels

    def channel(self, name: str) -> QDac2Channel:
        return self._qdac.channel(self._channels[self._contacts[name]])

    def virtual_voltage(self, contact: str) -> float:
        """
        Args:
            contact (str): Name of contact

        Returns:
            float: Voltage before correction
        """
        index = self._contact_index(contact)
        return self._virtual_voltages[index]

    def actual_voltages(self) -> Sequence[float]:
        """
        Returns:
            Sequence[float]: Corrected voltages for all contacts
        """
        vs = np.matmul(self._correction, self._virtual_voltages)
        if self._qdac._round_off:
            vs = np.round(vs, self._qdac._round_off)
        return list(vs)

    def get_trigger_by_name(self, name: str) -> QDac2Trigger_Context:
        """
        Args:
            name (str): Name of trigger

        Returns:
            QDac2Trigger_Context: Trigger context manager
        """
        try:
            return self._internal_triggers[name]
        except KeyError:
            print(f"Internal triggers: {list(self._internal_triggers.keys())}")
            raise

    def _all_channels_as_suffix(self) -> str:
        channels_str = ints_to_comma_separated_list(self.channel_numbers)
        return f"(@{channels_str})"

    def currents_A(self, nplc: int = 1, current_range: str = "low") -> Sequence[float]:
        """Measure currents on all contacts

        Args:
            nplc (int, optional): Number of powerline cycles to average over
            current_range (str, optional): Current range (default low)
        """
        channels_suffix = self._all_channels_as_suffix()
        self._qdac.write(f"sens:rang {current_range},{channels_suffix}")
        self._qdac.write(f"sens:nplc {nplc},{channels_suffix}")
        # Discard first reading because of possible output-capacitor effects, etc
        slowest_line_freq_Hz = 50
        sleep_s(1 / slowest_line_freq_Hz)
        self._qdac.ask(f"read? {channels_suffix}")
        # Then make a proper reading
        sleep_s((nplc + 1) / slowest_line_freq_Hz)
        currents = self._qdac.ask(f"read? {channels_suffix}")
        return comma_sequence_to_list_of_floats(currents)

    def virtual_sweep(
        self,
        contact: str,
        voltages: Sequence[float],
        start_sweep_trigger: Optional[str] = None,
        step_time_s: float = 1e-5,
        step_trigger: Optional[str] = None,
        repetitions: int = 1,
    ) -> Virtual_Sweep_Context:
        """Sweep a contact to create a 1D sweep

        Args:
            contact (str): Name of sweeping contact
            voltages (Sequence[float]): Virtual sweep voltages
            outer_contact (str): Name of slow-changing (outer) contact
            start_sweep_trigger (None, optional): Trigger that starts sweep
            step_time_s (float, optional): Delay between voltage changes
            step_trigger (None, optional): Trigger that marks each step
            repetitions (int, Optional): Number of back-and-forth sweeps, or -1 for infinite

        Returns:
            Virtual_Sweep_Context: context manager
        """
        sweep = self._calculate_1d_values(contact, voltages)
        return Virtual_Sweep_Context(self, sweep, start_sweep_trigger, step_time_s, step_trigger, repetitions)

    def _calculate_1d_values(self, contact: str, voltages: Sequence[float]) -> np.ndarray:
        original_voltage = self.virtual_voltage(contact)
        index = self._contact_index(contact)
        sweep = list()
        for v in voltages:
            self._virtual_voltages[index] = v
            sweep.append(self.actual_voltages())
        self._virtual_voltages[index] = original_voltage
        return np.array(sweep)

    def virtual_sweep2d(
        self,
        inner_contact: str,
        inner_voltages: Sequence[float],
        outer_contact: str,
        outer_voltages: Sequence[float],
        start_sweep_trigger: Optional[str] = None,
        inner_step_time_s: float = 1e-5,
        inner_step_trigger: Optional[str] = None,
        repetitions: int = 1,
    ) -> Virtual_Sweep_Context:
        """Sweep two contacts to create a 2D sweep

        Args:
            inner_contact (str): Name of fast-changing (inner) contact
            inner_voltages (Sequence[float]): Inner contact virtual voltages
            outer_contact (str): Name of slow-changing (outer) contact
            outer_voltages (Sequence[float]): Outer contact virtual voltages
            start_sweep_trigger (None, optional): Trigger that starts sweep
            inner_step_time_s (float, optional): Delay between voltage changes
            inner_step_trigger (None, optional): Trigger that marks each step
            repetitions (int, Optional): Number of back-and-forth sweeps, or -1 for infinite

        Returns:
            Virtual_Sweep_Context: context manager
        """
        sweep = self._calculate_2d_values(inner_contact, inner_voltages, outer_contact, outer_voltages)
        return Virtual_Sweep_Context(
            self, sweep, start_sweep_trigger, inner_step_time_s, inner_step_trigger, repetitions
        )

    def _calculate_2d_values(
        self, inner_contact: str, inner_voltages: Sequence[float], outer_contact: str, outer_voltages: Sequence[float]
    ) -> np.ndarray:
        original_fast_voltage = self.virtual_voltage(inner_contact)
        original_slow_voltage = self.virtual_voltage(outer_contact)
        outer_index = self._contact_index(outer_contact)
        inner_index = self._contact_index(inner_contact)
        sweep = list()
        for slow_V in outer_voltages:
            self._virtual_voltages[outer_index] = slow_V
            for fast_V in inner_voltages:
                self._virtual_voltages[inner_index] = fast_V
                sweep.append(self.actual_voltages())
        self._virtual_voltages[inner_index] = original_fast_voltage
        self._virtual_voltages[outer_index] = original_slow_voltage
        return np.array(sweep)

    def virtual_detune(
        self,
        contacts: Sequence[str],
        start_V: Sequence[float],
        end_V: Sequence[float],
        steps: int,
        start_trigger: Optional[str] = None,
        step_time_s: float = 1e-5,
        step_trigger: Optional[str] = None,
        repetitions: int = 1,
    ) -> Virtual_Sweep_Context:
        """Sweep any number of contacts linearly from one set of values to another set of values

        Args:
            contacts (Sequence[str]): contacts involved in sweep
            start_V (Sequence[float]): First-extreme values
            end_V (Sequence[float]): Second-extreme values
            steps (int): Number of steps between extremes
            start_trigger (None, optional): Trigger that starts sweep
            step_time_s (float, Optional): Seconds between each step
            step_trigger (None, optional): Trigger that marks each step
            repetitions (int, Optional): Number of back-and-forth sweeps, or -1 for infinite
        """
        self._check_same_lengths(contacts, start_V, end_V)
        sweep = self._calculate_detune_values(contacts, start_V, end_V, steps)
        return Virtual_Sweep_Context(self, sweep, start_trigger, step_time_s, step_trigger, repetitions)

    @staticmethod
    def _check_same_lengths(contacts, start_V, end_V) -> None:
        n_contacts = len(contacts)
        if n_contacts != len(start_V):
            raise ValueError(f"There must be exactly one voltage per contact: {start_V}")
        if n_contacts != len(end_V):
            raise ValueError(f"There must be exactly one voltage per contact: {end_V}")

    def _calculate_detune_values(
        self, contacts: Sequence[str], start_V: Sequence[float], end_V: Sequence[float], steps: int
    ):
        original_voltages = [self.virtual_voltage(contact) for contact in contacts]
        indices = [self._contact_index(contact) for contact in contacts]
        sweep = list()
        forward_V = [forward_and_back(start_V[i], end_V[i], steps) for i in range(len(contacts))]
        for voltages in zip(*forward_V):
            for index, voltage in zip(indices, voltages):
                self._virtual_voltages[index] = voltage
            sweep.append(self.actual_voltages())
        for index, voltage in zip(indices, original_voltages):
            self._virtual_voltages[index] = voltage
        return np.array(sweep)

    def leakage(self, modulation_V: float, nplc: int = 2) -> np.ndarray:
        """Run a simple leakage test between the contacts

        Each contact is changed in turn and the resulting change in current from
        steady-state is recorded.  The resulting resistance matrix is calculated
        as modulation_voltage divided by current_change.

        Args:
            modulation_V (float): Virtual voltage added to each contact
            nplc (int, Optional): Powerline cycles to wait for each measurement

        Returns:
            ndarray: contact-to-contact resistance in Ohms
        """
        steady_state_A, currents_matrix = self._leakage_currents(modulation_V, nplc, "low")
        with np.errstate(divide="ignore"):
            return np.abs(modulation_V / diff_matrix(steady_state_A, currents_matrix))

    def _leakage_currents(
        self, modulation_V: float, nplc: int, current_range: str
    ) -> tuple[Sequence[float], Sequence[Sequence[float]]]:
        steady_state_A = self.currents_A(nplc, "low")
        currents_matrix = list()
        for index, channel_nr in enumerate(self.channel_numbers):
            original_V = self._virtual_voltages[index]
            self._effectuate_virtual_voltage(index, original_V + modulation_V)
            currents = self.currents_A(nplc, current_range)
            self._effectuate_virtual_voltage(index, original_V)
            currents_matrix.append(currents)
        return steady_state_A, currents_matrix

    def _contact_index(self, contact: str) -> int:
        return self._contacts[contact]

    def _allocate_triggers(
        self, internal_triggers: Optional[Sequence[str]], output_triggers: Optional[dict[str, int]]
    ) -> None:
        self._internal_triggers: dict[str, QDac2Trigger_Context] = dict()
        self._allocate_internal_triggers(internal_triggers)
        self._allocate_external_triggers(output_triggers)

    def _allocate_external_triggers(self, output_triggers: Optional[dict[str, int]]) -> None:
        self._external_triggers = dict()
        if not output_triggers:
            return
        for name, port in output_triggers.items():
            self._external_triggers[name] = port
            trigger = self._qdac.allocate_trigger()
            self._qdac.connect_external_trigger(port, trigger)
            self._internal_triggers[name] = trigger

    def _free_triggers(self) -> None:
        for trigger in self._internal_triggers.values():
            self._qdac.free_trigger(trigger)


def forward_and_back(start: float, end: float, steps: int):
    forward = np.linspace(start, end, steps)
    backward = np.flip(forward)[1:][:-1]
    back_and_forth = itertools.chain(forward, backward)
    return back_and_forth


class QDac2(VisaInstrument):

    def __init__(self, name: str, address: str, **kwargs) -> None:
        """Connect to a QDAC-II

        Args:
            name (str): Name for instrument
            address (str): Visa identification string
            **kwargs: additional argument to the Visa driver
        """
        self._check_instrument_name(name)
        super().__init__(name, address, terminator="\n", **kwargs)
        self._set_up_serial()
        self._set_up_debug_settings()
        self._set_up_channels()
        self._set_up_external_triggers()
        self._set_up_internal_triggers()
        self._set_up_simple_functions()
        self.connect_message()
        self._check_for_wrong_model()
        self._check_for_incompatiable_firmware()
        self._set_up_manual_triggers()

    def n_channels(self) -> int:
        """
        Returns:
            int: Number of channels
        """
        return len(self.submodules["channels"])

    def channel(self, ch: int) -> QDac2Channel:
        """
        Args:
            ch (int): Channel number

        Returns:
            QDac2Channel: Visa representation of the channel
        """
        return getattr(self, f"ch{ch:02}")

    @staticmethod
    def n_triggers() -> int:
        """
        Returns:
            int: Number of internal triggers
        """
        return 14

    @staticmethod
    def n_external_inputs() -> int:
        """
        Returns:
            int: Number of external input triggers
        """
        return 4

    def n_external_outputs(self) -> int:
        """
        Returns:
            int: Number of external output triggers
        """
        return len(self.submodules["external_triggers"])

    def allocate_trigger(self) -> QDac2Trigger_Context:
        """Allocate an internal trigger

        Does not have any effect on the instrument, only the driver.

        Returns:
            QDac2Trigger_Context: Context manager

        Raises:
            ValueError: no free triggers
        """
        try:
            number = self._internal_triggers.pop()
        except KeyError:
            raise ValueError("no free internal triggers")
        return QDac2Trigger_Context(self, number)

    def free_trigger(self, trigger: QDac2Trigger_Context) -> None:
        """Free an internal trigger

        Does not have any effect on the instrument, only the driver.

        Args:
            trigger (QDac2Trigger_Context): trigger to free
        """
        internal = _trigger_context_to_value(trigger)
        self._internal_triggers.add(internal)

    def free_all_triggers(self) -> None:
        """Free all an internal triggers

        Does not have any effect on the instrument, only the driver.
        """
        self._set_up_internal_triggers()

    def connect_external_trigger(self, port: int, trigger: QDac2Trigger_Context, width_s: float = 1e-6) -> None:
        """Route internal trigger to external trigger

        Args:
            port (int): External output trigger number
            trigger (QDac2Trigger_Context): Internal trigger
            width_s (float, optional): Output trigger width in seconds (default 1ms)
        """
        internal = _trigger_context_to_value(trigger)
        self.write(f"outp:trig{port}:sour int{internal}")
        self.write(f"outp:trig{port}:widt {width_s}")

    def reset(self) -> None:
        self.write("*rst")
        sleep_s(5)

    def errors(self) -> str:
        """Retrieve and clear all previous errors

        Returns:
            str: Comma separated list of errors or '0, "No error"'
        """
        return self.ask("syst:err:all?")

    def error(self) -> str:
        """Retrieve next error

        Returns:
            str: The next error or '0, "No error"'
        """
        return self.ask("syst:err?")

    def n_errors(self) -> int:
        """Peek at number of previous errors

        Returns:
            int: Number of errors
        """
        return int(self.ask("syst:err:coun?"))

    def start_all(self) -> None:
        """Trigger the global SCPI bus (``*TRG``)

        All generators, that have not been explicitly set to trigger on an
        internal or external trigger, will be started.
        """
        self.write("*trg")

    def remove_traces(self) -> None:
        """Delete all trace definitions from the instrument

        This means that all AWGs loose their data.
        """
        self.write("trac:rem:all")

    def traces(self) -> Sequence[str]:
        """List all defined traces

        Returns:
            Sequence[str]: trace names
        """
        return comma_sequence_to_list(self.ask("trac:cat?"))

    def allocate_trace(self, name: str, size: int) -> Trace_Context:
        """Reserve memory for a new trace

        Args:
            name (str): Name of new trace
            size (int): Number of voltage values in the trace

        Returns:
            Trace_Context: context manager
        """
        return Trace_Context(self, name, size)

    def mac(self) -> str:
        """
        Returns:
            str: Media Access Control (MAC) address of the instrument
        """
        mac = self.ask("syst:comm:lan:mac?")
        return f"{mac[1:3]}-{mac[3:5]}-{mac[5:7]}-{mac[7:9]}-{mac[9:11]}" f"-{mac[11:13]}"

    def arrange(
        self,
        contacts: dict[str, int],
        output_triggers: Optional[dict[str, int]] = None,
        internal_triggers: Optional[Sequence[str]] = None,
    ) -> Arrangement_Context:
        """An arrangement of contacts and triggers for virtual gates

        Each contact corresponds to a particular output channel.  Each
        output_trigger corresponds to a particular external output trigger.
        Each internal_trigger will be allocated from the pool of internal
        triggers, and can later be used for synchronisation.  After
        initialisation of the arrangement, contacts and triggers can only be
        referred to by name.

        The voltages that will appear on each contact depends not only on the
        specified virtual voltage, but also on a correction matrix.  Initially,
        the contacts are assumed to not influence each other, which means that
        the correction matrix is the identity matrix, ie. the row for
        each contact has a value of [0, ..., 0, 1, 0, ..., 0].

        Args:
            contacts (Dict[str, int]): Name/channel pairs
            output_triggers (Sequence[Tuple[str,int]], optional): Name/number pairs of output triggers
            internal_triggers (Sequence[str], optional): List of names of internal triggers to allocate

        Returns:
            Arrangement_Context: context manager
        """
        return Arrangement_Context(self, contacts, output_triggers, internal_triggers)

    # -----------------------------------------------------------------------
    # Instrument-wide functions
    # -----------------------------------------------------------------------

    # -----------------------------------------------------------------------
    # Debugging and testing

    def start_recording_scpi(self) -> None:
        """Record all SCPI commands sent to the instrument

        Any previous recordings are removed.  To inspect the SCPI commands sent
        to the instrument, call get_recorded_scpi_commands().
        """
        self._scpi_sent: list[str] = list()
        self._record_commands = True

    def get_recorded_scpi_commands(self) -> list[str]:
        """
        Returns:
            Sequence[str]: SCPI commands sent to the instrument
        """
        commands = self._scpi_sent
        self._scpi_sent = list()
        return commands

    def clear(self) -> None:
        """Reset the VISA message queue of the instrument"""
        self.visa_handle.clear()

    def clear_read_queue(self) -> Sequence[str]:
        """Flush the VISA message queue of the instrument

        Takes at least _message_flush_timeout_ms to carry out.

        Returns:
            Sequence[str]: Messages lingering in queue
        """
        lingering = list()
        original_timeout = self.visa_handle.timeout
        self.visa_handle.timeout = self._message_flush_timeout_ms
        while True:
            try:
                message = self.visa_handle.read()
            except VisaIOError:
                break
            else:
                lingering.append(message)
        self.visa_handle.timeout = original_timeout
        return lingering

    # -----------------------------------------------------------------------
    # Override communication methods to make it possible to record the
    # communication with the instrument.

    def write(self, cmd: str) -> None:
        """Send SCPI command to instrument

        Args:
            cmd (str): SCPI command
        """
        if self._record_commands:
            self._scpi_sent.append(cmd)
        super().write(cmd)

    def ask(self, cmd: str) -> str:
        """Send SCPI query to instrument

        Args:
            cmd (str): SCPI query

        Returns:
            str: SCPI answer
        """
        if self._record_commands:
            self._scpi_sent.append(cmd)
        answer = super().ask(cmd)
        return answer

    def write_floats(self, cmd: str, values: Sequence[float]) -> None:
        """Append a list of values to a SCPI command

        By default, the values are IEEE binary encoded.

        Remember to include separating space in command if needed.
        """
        if self._no_binary_values:
            compiled = f"{cmd}{floats_to_comma_separated_list(values)}"
            if self._record_commands:
                self._scpi_sent.append(compiled)
            return super().write(compiled)
        if self._record_commands:
            self._scpi_sent.append(f"{cmd}{floats_to_comma_separated_list(values)}")
        self.visa_handle.write_binary_values(cmd, values)

    # -----------------------------------------------------------------------

    def _set_up_debug_settings(self) -> None:
        self._record_commands = False
        self._scpi_sent = list()
        self._message_flush_timeout_ms = 1
        self._round_off = None
        self._no_binary_values = False

    def _set_up_serial(self) -> None:
        # No harm in setting the speed even if the connection is not serial.
        self.visa_handle.baud_rate = 921600  # type: ignore

    def _check_for_wrong_model(self) -> None:
        model = self.IDN()["model"]
        if model != "QDAC-II":
            raise ValueError(f"Unknown model {model}. Are you using the right" " driver for your instrument?")

    def _check_for_incompatiable_firmware(self) -> None:
        # Only compare the firmware, not the FPGA version
        firmware = split_version_string_into_components(self.IDN()["firmware"])[1]
        least_compatible_fw = "0.17.5"
        if parse(firmware) < parse(least_compatible_fw):
            raise ValueError(f"Incompatible firmware {firmware}. You need at " f"least {least_compatible_fw}")

    def _set_up_channels(self) -> None:
        channels = ChannelList(self, "Channels", QDac2Channel, snapshotable=False)
        for i in range(1, 24 + 1):
            name = f"ch{i:02}"
            channel = QDac2Channel(self, name, i)
            self.add_submodule(name, channel)
            channels.append(channel)
        channels.lock()
        self.add_submodule("channels", channels)

    def _set_up_external_triggers(self) -> None:
        triggers = ChannelList(self, "Channels", QDac2ExternalTrigger, snapshotable=False)
        for i in range(1, 5 + 1):
            name = f"ext{i}"
            trigger = QDac2ExternalTrigger(self, name, i)
            self.add_submodule(name, trigger)
            triggers.append(trigger)
        triggers.lock()
        self.add_submodule("external_triggers", triggers)

    def _set_up_internal_triggers(self) -> None:
        # A set of the available internal triggers
        self._internal_triggers = set(range(1, self.n_triggers() + 1))

    def _set_up_manual_triggers(self) -> None:
        self.add_parameter(
            name="trigger",
            # Manually trigger event
            set_parser=_trigger_context_to_value,
            set_cmd="tint {}",
        )

    def _set_up_simple_functions(self) -> None:
        self.add_function("abort", call_cmd="abor")

    def _check_instrument_name(self, name: str) -> None:
        if name.isidentifier():
            return
        raise ValueError(
            f'Instrument name "{name}" is incompatible with QCoDeS parameter '
            "generation (no spaces, punctuation, prepended numbers, etc)"
        )



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman

from .generic_measurement import (
    Generic_1D_Hysteresis_buffered,
    Generic_1D_parallel_asymm_Sweep,
    Generic_1D_parallel_Sweep,
    Generic_1D_Sweep,
    Generic_1D_Sweep_buffered,
    Generic_2D_Sweep_buffered,
    Generic_nD_Sweep,
    Generic_Pulsed_Measurement,
    Generic_Pulsed_Repeated_Measurement,
    Timetrace,
    Timetrace_buffered,
    Timetrace_with_sweeps,
    Timetrace_with_Sweeps_buffered,
)

try:
    from .spectrometer import Measure_Spectrum
except ModuleNotFoundError:
    # Only relevant if you want to use spectrometer.
    # Requires access to Bluhm Group GitLab
    pass
except ImportError:
    pass

__all__ = [
    "Generic_1D_Sweep",
    "Generic_1D_Sweep_buffered",
    "Generic_nD_Sweep",
    "Generic_1D_parallel_Sweep",
    "Generic_1D_Hysteresis_buffered",
    "Generic_1D_parallel_asymm_Sweep",
    "Generic_2D_Sweep_buffered",
    "Generic_Pulsed_Measurement",
    "Generic_Pulsed_Repeated_Measurement",
    "Timetrace",
    "Timetrace_with_sweeps",
    "Timetrace_buffered",
    "Timetrace_with_Sweeps_buffered",
    "Measure_Spectrum",
]



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Sionludi Lab
# - Till Huckeman


"""
Loading data from Database
"""
from __future__ import annotations

from os import path

import numpy as np
import qcodes as qc
from qcodes.dataset.data_set import DataSet
from qcodes.dataset.plotting import plot_dataset

from qumada.utils.browsefiles import browsefiles


# %%
def flatten_list(lst: list) -> list:
    """
    Flattens nested lists
    """
    results = []

    def rec(sublist, results):
        for entry in sublist:
            if isinstance(entry, list):
                rec(entry, results)
            else:
                results.append(entry)

    rec(lst, results)
    return results


# %%
def load_db(filepath: str | None = None) -> None:
    """
    Loads or creates the database.

    Parameters
    ----------
    filepath : str, optional
        Provide the path to the DB here if you want to skip the file explorer.
        The default is None.

    Returns
    -------
    None.
    #TODO: Checks only whether provided path is a file but not which type.
    """
    if not filepath:
        filetypes = (("DB Files", "*.db*"), ("All files", "*.*"))
        filepath = browsefiles(filetypes=filetypes)
        if filepath == "":
            return None
        try:
            qc.initialise_or_create_database_at(filepath)
            return None
        except Exception:
            print("Please provide a valid path")
            return load_db(None)
    elif not path.isfile(filepath):
        try:
            qc.initialise_or_create_database_at(filepath)
            print("Created new Database")
            return None
        except Exception as e:
            print("Please provide a valid path")
            raise e
    else:
        try:
            qc.initialise_or_create_database_at(filepath)
        except Exception:
            print("The file you want to load is no valid db file!")
            return load_db(None)


# %%
def list_sample_names() -> set[str]:
    """
    Lists all sample names that appear in the database
    """
    name_set = {measurement.sample_name for measurement in qc.experiments()}
    return name_set


# %%
def _pick_sample_name() -> str:
    """
    Lists all samples and allows the user to pick one.
    Returns String with sample name
    """
    samples = list(list_sample_names())
    print("Please choose a sample:")
    for idx, sample in enumerate(samples):
        print(f"{idx}: {sample}")
    while True:
        try:
            chosen = samples[int(input("Enter sample number: "))]
            return chosen
        except Exception:
            print("Please chose a valid entry")


# %%
def list_measurements_for_sample(sample_name: str | None = None) -> list[qc.DataSet]:
    """
    Lists all measurements done with a certain sample in the console.
    If no sample is provided it helps you to find one.
    Returns list with all datasets belonging to the sample specified.
    Parameters
    ----------
    sample_name : str, optional
        In case you do not want to do this via user input. The default is None.

    Returns
    -------
    list
    """
    if not sample_name:
        sample_name = _pick_sample_name()
    try:
        qc.load_by_run_spec(sample_name=sample_name)
    except NameError:
        pass
    return _list_measurements_for_sample(sample_name)


# %%
def _list_measurements_for_sample(sample_name: str | None = None) -> list[qc.DataSet]:
    """
    Returns flattened list containing all datasets belonging to the sample specified
    """
    datasets = []
    for experiment in qc.experiments():
        if experiment.sample_name == sample_name:
            datasets.append([*(dataset for dataset in experiment.data_sets())])
    return flatten_list(datasets)


# %%
def _flatten_experiment_container() -> list[qc.DataSet]:
    """
    Returns flattened list of all datasets in the currently loaded .db
    """
    datasets = [*(experiment.data_sets() for experiment in qc.experiments())]
    return flatten_list(datasets)


# %%


def pick_measurement(sample_name: str = None, preview_dialogue=True):
    """
    Returns a measurement of your choice, plots it if you want.
    Interactive, if no sample_name is provided.
    """

    measurements = list_measurements_for_sample(sample_name=sample_name)
    for idx, measurement in enumerate(measurements):
        print(f"{idx} (Run ID {measurement.run_id}) : {measurement.name}")
    chosen = int(input("Please choose a measurement: "))
    chosen_measurement = measurements[int(chosen)]
    if preview_dialogue:
        preview = str(input("Do you want to see a plot? (Y/N) "))
        if str.lower(preview) == "y":
            plot_dataset(chosen_measurement)
    return chosen_measurement


# %%


def pick_measurements(sample_name: str = None, preview_dialogue=False, measurement_list=None):
    """
    Returns a measurement of your choice, plots it if you want.
    Interactive, if no sample_name is provided.
    """
    if not measurement_list:
        measurement_list = list()
    measurements = list_measurements_for_sample(sample_name=sample_name)
    for idx, measurement in enumerate(measurements):
        print(f"{idx} (Run ID {measurement.run_id}) : {measurement.name}")
    while True:
        chosen = input("Please choose a measurement: ")
        if chosen == "f":
            return measurement_list
        if chosen == "s":
            return pick_measurements(preview_dialogue=preview_dialogue, measurement_list=measurement_list)
        chosen = int(chosen)
        measurement_list.append(measurements[int(chosen)])
        print("Please enter 'f' when your are finished or 's' if you want to add measurements of another sample")


# %%


def plot_data(sample_name: str = None):
    """
    Simple plotting of datasets from the QCoDeS DB.
    """
    dataset = pick_measurement(sample_name=sample_name, preview_dialogue=False)
    independend_parameters = list()
    dependend_parameters = list()
    for parameter in dataset.get_parameters():
        if len(parameter._depends_on) == 0:
            independend_parameters.append(parameter)
        else:
            dependend_parameters.append(parameter)
    print("Which parameter do you want to plot?")
    for idx, parameter in enumerate(dependend_parameters):
        print(f"{idx} : {parameter.label}")
    plot_param_numbers = input("Please enter the numbers of the parameters you want to plot, separated by blank")
    plot_params = list()
    for param in plot_param_numbers:
        plot_params.append(dependend_parameters[int(param)].name)
    print(plot_params)
    for param in plot_params:
        y_data = dataset.get_parameter_data(param)[param][param]
        print(dataset.get_parameter_data(param)[param])
        x_data = dataset.get_parameter_data(param)[param][dependend_parameters[0].name]

        print(y_data)
        print(x_data)


# %%
def get_parameter_data(dataset=None, parameter_name=None, **kwargs):
    """
    Gets you the data for a chosen dependent parameter and the data of the first (!)
    parameter it depends on
    #TODO: Support for independent parameters
    """
    if not dataset:
        dataset = pick_measurement()
    if not isinstance(dataset, DataSet):
        print("Dataset has to be of type DataSet")
        return None
    if not parameter_name:
        for idx, parameter in enumerate(dataset.dependent_parameters):
            print(f"{idx} : {parameter.label} / {parameter.name}")
        chosen_param_num = int(input("Please enter number: "))
        chosen_param = dataset.dependent_parameters[chosen_param_num]
        parameter_name = chosen_param.name
    else:
        chosen_param = dataset.paramspecs[parameter_name]
    independent_param = dataset.paramspecs[parameter_name]._depends_on
    # if not isinstance(independent_param, str):
    #     independent_param = independent_param[0]
    params = (*independent_param, parameter_name)
    labels = (*(dataset.paramspecs[i_p].label for i_p in independent_param), chosen_param.label)
    units = (*(dataset.paramspecs[i_p].unit for i_p in independent_param), chosen_param.unit)
    data = (
        *tuple(dataset.get_parameter_data(parameter_name)[parameter_name][param] for param in independent_param),
        dataset.get_parameter_data(parameter_name)[parameter_name][parameter_name],
    )
    return zip(params, data, units, labels)


# %%
def separate_up_down(x_data, y_data):
    grad = np.gradient(x_data)
    curr_sign = np.sign(grad[0])
    data_list_x = list()
    data_list_y = list()
    direction = list()
    direction.append(curr_sign)
    start_helper = 0
    for i in range(0, len(grad)):
        if np.sign(grad[i]) != curr_sign:
            data_list_x.append(x_data[start_helper:i])
            data_list_y.append(y_data[start_helper:i])
            start_helper = i + 1
            curr_sign = np.sign(grad[i])
            direction.append(curr_sign)
    data_list_x.append(x_data[start_helper : len(grad)])
    data_list_y.append(y_data[start_helper : len(grad)])
    if len(direction) == 0:
        direction.append(1)
    return data_list_x, data_list_y, direction



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman


from __future__ import annotations

from qcodes.parameters import Parameter

from qumada.instrument.custom_drivers.Dummies.dummy_dac import DummyDac
from qumada.instrument.mapping import DUMMY_DAC_MAPPING
from qumada.instrument.mapping.base import InstrumentMapping


class DummyDacMapping(InstrumentMapping):
    def __init__(self):
        super().__init__(DUMMY_DAC_MAPPING)

    def ramp(
        self,
        parameters: list[Parameter],
        *,
        start_values: list[float] | None = None,
        end_values: list[float],
        ramp_time: float,
        **kwargs,
    ) -> None:
        num_points = kwargs.get("num_points", 100 * ramp_time)
        assert len(parameters) == len(end_values)
        if start_values is not None:
            assert len(parameters) == len(start_values)

        if len(parameters) > 4:
            raise Exception("Maximum length of rampable parameters currently is 4.")
        # TODO: Test delay when ramping multiple parameters in parallel.
        # TODO: Add Trigger option?
        # check, if all parameters are from the same instrument
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception("Parameters are from more than one instrument. This would lead to non synchronized ramps.")

        instrument: DummyDac = instruments.pop()
        assert isinstance(instrument, DummyDac)

        if not start_values:
            start_values = [param.get() for param in parameters]

        instrument._triggered_ramp_channels(
            [param._instrument for param in parameters], start_values, end_values, ramp_time, num_points
        )

    def pulse(
        self,
        parameters: list[Parameter],
        *,
        setpoints: list[list[float]],
        delay: float,
        sync_trigger=None,
        **kwargs,
    ) -> None:
        assert len(parameters) == len(setpoints)
        num_points = len(setpoints[0])
        for setpoint in setpoints:
            assert len(setpoint) == num_points

        duration = num_points * delay
        if len(parameters) > 4:
            raise Exception("Maximum length of pulsable parameters currently is 4.")
        # TODO: Test delay when ramping multiple parameters in parallel.
        # TODO: Add Trigger option?
        # check, if all parameters are from the same instrument
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception("Parameters are from more than one instrument. This would lead to non synchronized ramps.")

        instrument: DummyDac = instruments.pop()
        assert isinstance(instrument, DummyDac)
        instrument._triggered_pulse_channels([param._instrument for param in parameters], setpoints, duration)

    def setup_trigger_in():
        pass



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman


def get_station_instr(station):
    """
    Lists all instruments listed in a qcodes station config yaml.
    Can be used to create instances of instruments after loading the station.
    """
    result = []
    for elem in station.__dict__["_added_methods"]:
        result.append(elem.strip("load_"))
    return result


def instance_instr(station):
    """
    Creates instances of all instruments listed in qcodes stations qconfig yaml
    and puts them into a list. Not the most elegant way but might be handy for
    automation...
    """
    instruments = []
    for elem in get_station_instr(station):
        instruments.append(station.load_instrument(elem))
    return instruments
